import pdb
import warnings
import numpy as np
from numpy import NaN, Inf
import os, glob, csv, time, scipy
import pandas as pd
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
import obspy
import sys
from scipy.io import wavfile
import matplotlib.pyplot as plt

_debug = False

class EmptyRawFile(Exception):
    """Raised when the input file is empty"""
    pass
class CorruptRawFile(Exception):
    """Raised when the input file cannot be read for some reason"""
    pass

class CorruptRawFileNoGPS(CorruptRawFile):
    """Raised when the input file does not contain any GPS data"""
    pass

class CorruptRawFileInadequateGPS(CorruptRawFile):
    """Raised when the input file does not contain any GPS data"""
    pass

class MissingRawFiles(Exception):
    """Raised when no input files are readable"""
    pass


def _breakpoint():
    if _debug: # skip if we aren't in debug mode
        pdb.set_trace()
        
#####################
def convert(rawpath = '.', convertedpath = 'converted', metadatapath = 'metadata', \
            metadatafile = '', gpspath = 'gps', gpsfile = '', t1 = -Inf, t2 = Inf, nums = NaN, \
            SN = '', bitweight = NaN, units = 'Pa', time_adjustment = 0, blockdays = 1, \
            file_length_hour = 1, station = '', network = '', location = '', output_format = 'MSEED'):
    """
    Read raw Gem files, interpolate them, and write output files in miniSEED or SAC format.

    Parameters
    ----------
    rawpath : str, default '.' 
        Path of folder containing raw Gem data files to read.

    convertedpath : str, 'converted'
        Path of folder where converted data files should be written. If this folder does 
        not exist, convert will try to create it.
    
    metadatapath : str, 'metadata'
        Path of folder where metadata files should be written. If this folder does 
        not exist, convert will try to create it.
    
    metadatafile : str
        File name for metadata file. Default is of the form 'XXXmetadata_NNN.txt', where
        'XXX' is the Gem's serial number and 'NNN' is the file index (000 at first, incrementing by one for each subsequent calculation).

    gpspath : str, 'gps'
        Path of folder where gps files should be written. If this folder does 
        not exist, convert will try to create it.
    
    gpsfile : str
        File name for gps file. Default is of the form 'XXXgps_NNN.txt', where
        'XXX' is the Gem's serial number and 'NNN' is the file index (000 at first, incrementing by one for each subsequent calculation).
    
    t1 : float or obspy.UTCDateTime, default -np.inf
        Start time for the conversion. If float, the number of seconds 
        since the epoch (1970-01-01 00:00:00 UTC). 

    t2 : float or obspy.UTCDateTime, default np.inf
        Stop time for the conversion. If float, the number of seconds 
        since the epoch (1970-01-01 00:00:00 UTC). 

    nums : list or np.array of integers
        Numbers of raw Gem files to read. By default, it reads all files
        in rawpath for the specified serial number.
    
    SN : str
        One Gem serial number to read and convert. Use a loop to convert
        multiple Gems.

    bitweight : float
        The value of each count when converting between counts and other 
        units (typically Pascals, possibly Volts). By default, it looks
        up the correct bitweight given the Gem's serial number and gain 
        configuration. Leave this blank unless the Gem has been modified
        in a way that changes the bitweight.

    units : str, default 'Pa'
        Desired output units. Options are 'Pa' (Pascals), 'V' (Volts), 
        or 'counts'.

    time_adjustment : float, default 0
        Amount to shift the sample times by in case of timing error, 
        possibly due to leap second issues.

    blockdays : float, default 1
        Number of days worth of data to read in and convert at a time.
        If you have memory issues when converting data, try setting this
        to a smaller value.

    file_length_hour : float, default 1
        Length of the output files in hours.
            
    station : str
        Name of the station (up to five characters) to assign to the 
        data. If not provided, uses the Gem's serial number as the 
        station ID.

    network : str
        Two-character name of the sensor network. Leaving this blank is
        normally fine in subsequent data processing.

    location : str
        Two-character location code for this Gem. Leaving this blank is 
        usually fine in subsequent data processing.
    
    output_format : str, default 'MSEED'
        Output file format. Currently, formats 'MSEED' and 'SAC' are 
        supported; 'WAV' is partly supported.

    Returns
    -------
    None, writes output files only (converted, metadata, and gps)

    Note
    ----
    All sample times involving the Gem (and most other passive 
    seismic/acoustic data) are in UTC; time zones are not supported.
    """
    file_length_sec = 3600 * float(file_length_hour)
    ## bitweight: leave blank to use default (considering Gem version, config, and units). This is preferred when using a standard Gem (R_g = 470 ohms)
    
    ## make sure the raw directory exists and has real data
    if not os.path.isdir(rawpath):
        raise MissingRawFiles('Raw directory ' + rawpath + ' does not exist')
    if len(glob.glob(rawpath + '/FILE' +'[0-9]'*4 + '.???')) == 0:
        raise MissingRawFiles('No data files found in directory ' + rawpath)
    try:
        SN = str(SN)
        int(SN) # make sure it's number-like
    except:
        raise Exception('Invalid serial number')
    if len(SN) != 3:
        raise Exception('Invalid serial number; SN type is length-'+ str(len(SN)) +' ' + str(type(SN)) + ', not length-3 str')
    
    ## make sure bitweight is a scalar
    if((type(nums) is int) or (type(nums) is float)):
        nums = np.array([nums])
    else:
        nums = np.array(nums)
    
    ## find a list of possibly eligible raw files 
    if (type(SN) is not str) or (len(SN) != 3): # check that SN is appropriately formatted...this should always happen
        fn = glob.glob(rawpath + '/FILE' +'[0-9]'*4 + '.???')
    else:
        fn = glob.glob(rawpath + '/FILE' +'[0-9]'*4 + '.' + SN) + \
            glob.glob(rawpath + '/FILE' +'[0-9]'*4 + '.TXT')
    
    ## filter out raw files whose serial numbers don't match SN
    fn_new = []
    corrupt_files_flag = False
    for file in fn:
        try:
            if _read_SN(file) == SN:
                fn_new.append(file)
        except:
            corrupt_files_flag = True
            pass # if we can't read the file's SN, it's corrupt and should be skipped
    fn = fn_new
    
    ## narrow the list of raw files if nums is provided, or calculate nums if not
    nums_from_fn = np.array([int(x[-8:-4]) for x in fn]) 
    if((len(nums) == 0) or np.isnan(nums[0])):
        nums = nums_from_fn
    else: # find the intersection between the available nums and the user-defined nums
        w = np.where([(i in nums) for i in nums_from_fn])[0]
        fn = [fn[i] for i in w]
        nums = nums_from_fn[w]

    ## Catch if rawpath doesn't contain any files from SN. This won't catch files ending in TXT.
    if len(nums) == 0:
        if corrupt_files_flag:
            raise CorruptRawFile('No non-corrupt data files for SN "' + SN + '" found in raw directory ' + rawpath)
        else:
            raise MissingRawFiles('No data files for SN "' + SN + '" found in raw directory ' + rawpath)

    ## start at the first file in 'nums'
    nums.sort()
    n1 = np.min(nums)
  
    ## read the first set of up to (24*blockdays) files
    L = _new_gem_var()
    while((L['data'].count() == 0) & (n1 <= max(nums))): ## read sets of files until we get one that isn't empty
        nums_block = nums[(nums >= n1) & (nums < (n1 + (12*blockdays)))] # files are 2 hours, so 12 files is 24 hours
        n1 = n1 + (12*blockdays) # increment file number counter
        try:
            L = read_gem(path = rawpath, nums = nums_block, SN = SN, network = network, station = station, location = location)
        except MissingRawFiles: # if the block has no files, keep searching
            continue
        except CorruptRawFile: # if the block has no files, keep searching
            continue
        except: # if there's any other problem, raise it
            raise
    if L['data'].count() == 0:
        raise MissingRawFiles(f'No non-corrupt raw files found in folder "{rawpath}"')
    p = L['data']
    
    ## if bitweight isn't set, use the default bitweight for the logger version, config, and units
    if(np.isnan(bitweight)):
        if(units == 'Pa'):
            bitweight = L['header']['bitweight_Pa'][0]
        elif(units == 'V'):
            bitweight = L['header']['bitweight_V'][0]
        elif (units == 'Counts') | (units == 'counts'):
            bitweight = 1
      
    ## if not specified, define t1 as the earliest integer-second time available
    if(np.isinf(float(t1))):
        p = L['data']
        p.merge()
        t1 = p[0].stats.starttime
        t1 = obspy.core.UTCDateTime(np.ceil(float(t1)))

    if(np.isinf(float(t2))):
        t2 = obspy.core.UTCDateTime.strptime('9999-12-31 23:59:59', '%Y-%m-%d %H:%M:%S') # timekeeping apocalypse
  
    wsn = 0
    while(len(SN) < 3): # take the first non-NA SN. This is important because there can be blank files in there.
        wsn = wsn+1
        SN = L['header']['SN'][wsn]
  
    ## set up the gps and metadata files. create directories if necessary
    if(len(gpsfile) == 0):
        if(not os.path.isdir(gpspath)):
            try:
                os.makedirs(gpspath) # makedirs vs mkdir means if gpspath = dir1/dir2, and dir1 doesn't exist, that dir1 will be created and then dir1/dir2 will be created
            except:
                print('Failed to make directory ' + gpspath)
                sys.exit(2)
        gpsfile = _make_filename(gpspath, SN, 'gps')

  
    if(len(metadatafile) == 0):
        if(not os.path.isdir(metadatapath)):
            try:
                os.makedirs(metadatapath)
            except:
                print('Failed to make directory ' + metadatapath)
                sys.exit(2)
        metadatafile = _make_filename(metadatapath, SN, 'metadata')
  
    ## if the converted directory does not exist, make it
    if(not os.path.isdir(convertedpath)):
        try:
            os.makedirs(convertedpath)
        except:
            print('Failed to make directory ' + convertedpath)
            sys.exit(2)
  
    ## start metadata and gps files
    metadata = L['metadata']   
    gps = L['gps']
    metadata.to_csv(metadatafile, index=False) ## change to metadata format. need to make ScanMnetadata compatible with both

    wgps = (gps['t'] > (t1 - 1)) 
    if(len(wgps) > 0):
        gps[wgps].to_csv(gpsfile, index=False)

    hour_to_write = max(t1, p[0].stats.starttime)
    hour_to_write = _write_hourlong_mseed(p, hour_to_write, file_length_sec, bitweight, convertedpath, output_format=output_format)
    
    ## read sets of (12*blockdays) files until all the files are converted
    while(True):
        ## check to see if we're done
        if(n1 > np.max(nums)):# & len(p) == 0):
            break # out of raw data to convert
        if((t1 > t2) & (not np.isnan(t1 > t2))):
            break # already converted the requested data
        ## load new data if necessary
        tt2 = min(t2, _trunc_UTCDateTime(t1, 86400*blockdays) + 86400*blockdays)
        while((p[-1].stats.endtime < tt2) & (n1 <= max(nums))):
            try:
                L = read_gem(path = rawpath, nums = nums[(nums >= n1) & (nums < (n1 + (12*blockdays)))], SN = SN)
            except MissingRawFiles: # this can happen if a block of empty files is encountered
                continue
            except CorruptRawFile: # if the block has no files, keep searching
                continue
            except: # especially for KeyboardInterrupt!
                raise
            finally:
                n1 = n1 + (12*blockdays) # increment file counter

            if(len(L['data']) == 0):
                continue # skip ahead if there aren't any readable data files here

            ## process newly-read data
            if(any(L['header'].SN != SN) | any(L['header'].SN.apply(len) == 0)):
                #_breakpoint()
                w = np.where((L['header'].SN != SN) | (L['header'].SN.apply(len) == 0))[0]
                #print('Wrong or missing serial number(s): ' + L['header'].SN[w] + ' : numbers ' + str(nums[np.logical_and(nums >= n1, nums < (n1 + (12*blockdays)))][w]))
                for i in w:
                    print('Problem with files, skipping: ' + L['header'].file[i])

            #breakpoint()
            if(len(L['data']) > 0):
                if (L['data'][0].stats.starttime - p[-1].stats.endtime) <= 0.031: # interpolate a gap up to 3 samples
                    L['data'] += p[-1]
                    L['data'].merge(fill_value = 'interpolate')
                ## merge all the data, regardless of whether we interpolated a gap above
                p = p + L['data']
                p.merge()
            #print(p)
                
            ## start metadata and gps files
            metadata = L['metadata']   
            gps = L['gps']
            metadata.to_csv(metadatafile, index=False, mode='a', header=False)
            wgps = (gps['t'] > (t1 - 1)) 
            ## update the gps file
            if(len(wgps) > 0):
                gps.to_csv(gpsfile, index=False, mode='a', header=False)
                
        ## run the conversion and write new converted files
        while((hour_to_write + file_length_sec) <= p[-1].stats.endtime):
            hour_to_write = _write_hourlong_mseed(p, hour_to_write, file_length_sec, bitweight, convertedpath, output_format=output_format)
            
        ## update start time to convert
        p.trim(hour_to_write, t2)
        t1 = _trunc_UTCDateTime(tt2+(86400*blockdays) + 1, 86400*blockdays)
    ## done reading new files. write what's left and end.
    #breakpoint()
    while((hour_to_write <= p[-1].stats.endtime) & (len(p) > 0)):
        hour_to_write = _write_hourlong_mseed(p, hour_to_write, file_length_sec, bitweight, convertedpath, output_format=output_format)
        p.trim(hour_to_write, t2)
        p = p.split()
        if(len(p) > 0):
            hour_to_write = p[0].stats.starttime
        else:
            break

Convert = convert # alias; v1.0.0
####################################

def _write_hourlong_mseed(p, hour_to_write, file_length_sec, bitweight, convertedpath, hour_end = np.nan, output_format='mseed'):
    eps = 1e-6
    if(np.isnan(hour_end)):
        hour_end = _trunc_UTCDateTime(hour_to_write, file_length_sec) + file_length_sec
    pp = p.slice(hour_to_write, hour_end - eps, nearest_sample = False) # nearest sample and eps avoid overlapping samples between files
    # Ideally, this would write mseed files with data gaps (masked arrays) if the stream here has
    # a gap. Unfortunately, those fail to write, so we have to write multiple files instead.
    pp = pp.split() 
    for tr in pp:
        tr.stats.calib = bitweight
        fn = _make_filename_converted(tr, output_format)
        if(len(tr) > 0):
            print(tr)
            if(output_format.lower() == 'wav'):
                write_wav(tr, filename = fn, path = convertedpath)
            else:
                tr.write(convertedpath +'/'+ fn, format = output_format, encoding=10) # encoding 10 is Steim 1
    hour_to_write = hour_end
    return hour_to_write

def write_wav(tr, filename = None, path = '.', time_format = '%Y-%m-%dT%H_%M_%S'):
    """
    Write a trace as a .wav file. This function is needed because obspy's 
    tr.write() method apparently does not handle wav files correctly.

    Parameters
    ----------
    tr : obspy.Trace()
        Trace containing data to be written to a .wav file.
    filename : str, default None
        Does not include path. If default None, creates a filename using time_format and station information.
    path : str, default '.'
        Path where file should be written.
    time_format : str, default '%Y-%m-%dT%H_%M_%S'
        In case filename is not provided explicitly, how to format the date/time
        in the output file name.
"""
    if filename is None:
        datetime_str = tr.stats.starttime.strftime(time_format)
        s = tr.stats
        station_str = '%s.%s.%s.%s' % (s.network, s.station, s.location, s.channel)
        filename = datetime_str + '.' + station_str + '.wav'
    ## having trouble with integer format, although it is allowed. force float
    ## format instead, which is probably more useful anyway. This needs to be
    ## centered about zero and scaled to fit in the -1 to 1 range.
    tr.data = tr.data.astype(float) - np.mean(tr.data)
    ref = np.abs(tr.data).max()
    if ref > 0:  ## prevent divide-by-zero
        tr.data /= ref
    ## sample rate must be an int
    if int(tr.stats.sampling_rate) != tr.stats.sampling_rate:
        raise TypeError('sample rate must be an integer')
    wavfile.write(path + '/' + filename, int(tr.stats.sampling_rate), tr.data)
    

def _trunc_UTCDateTime(x, n=86400):
    return obspy.core.UTCDateTime(int(float(x)/n)*n)#, origin='1970-01-01')

def _make_filename(dir, SN, dirtype):
    n = 0
    fn = dir + '/' + SN + dirtype + '_'+ f'{n:03}' + '.txt'
    while(os.path.exists(fn)):
        n = n + 1
        fn = dir + '/' + SN + dirtype + '_' + f'{n:03}' + '.txt'
    return fn


def _make_filename_converted(pp, output_format):
    t0 = pp.stats.starttime
    ## colons separating H:M:S would be more readable, but are not allowed in Windows filenames
    return f'{t0.year:04}' + '-' +f'{t0.month:02}' + '-' +f'{t0.day:02}' + 'T' + f'{t0.hour:02}' + '_' + f'{t0.minute:02}' + '_' + f'{t0.second:02}' + '.' + pp.id + '.' + output_format.lower()


##############################################################
##############################################################
def read_gem(path = 'raw', nums = np.arange(10000), SN = '', units = 'Pa', bitweight = np.NaN, bitweight_V = np.NaN, bitweight_Pa = np.NaN, verbose = True, network = '', station = '', location = '', return_debug_output = False, require_gps = True):
    """
    Read raw Gem files.

    Parameters
    ----------
    path : str, default '.' 
        Path of folder containing raw Gem data files to read.

    nums : list or np.array of integers
        Numbers of raw Gem files to read. By default, it reads all files
        in 'path' for the specified serial number.
    
    SN : str
        One Gem serial number to read. Use a loop to read multiple Gems.

    units : str, default 'Pa'
        Desired output units. Options are 'Pa' (Pascals), 'V' (Volts), 
        or 'counts'.

    bitweight : float
        The value of each count when converting between counts and other 
        units (typically Pascals, possibly Volts). By default, it looks
        up the correct bitweight given the Gem's serial number and gain 
        configuration. Leave this blank unless the Gem has been modified
        in a way that changes the bitweight.

    bitweight_V : float
        The value of each count when converting between counts and Volts. 
        By default, it looks up the correct bitweight given the Gem's 
        serial number and gain configuration. Leave blank unless the Gem
        has been modified in a way that changes the voltage bitweight.

    bitweight_Pa : float
        The value of each count when converting between counts and 
        Pascals. By default, it looks up the correct bitweight given the 
        Gem's serial number and gain configuration. Leave this blank 
        unless the Gem has been modified in a way that changes the 
        pressure bitweight.

    verbose : boolean, default True
        Whether to print verbose progress messages to the screen.

    network : str
        Two-character name of the sensor network. Leaving this blank is
        usually fine in subsequent data processing.

    station : str
        Name of the station (up to five characters) to assign to the 
        data. If not provided, uses the Gem's serial number as the 
        station ID.

    location : str
        Two-character location code for this Gem. Leaving this blank is 
        usually fine in subsequent data processing.

    return_debug_output : boolean, default False
        If True, return extra output to understand the internal state of the function.

    require_gps : boolean, default True
        If True, read files whether or not they have GPS data. Sample times will not be
        precise enough for array processing.

    Returns
    -------
    dict with keys:

        - data : obspy.Stream, infrasound data
        - header : pandas.dataframe, information on raw Gem data files
        - metadata : pandas.dataframe, state-of-health and other metadata time series
        - gps : pandas.dataframe, GPS timing and location values

    Note
    ----
    All sample times involving the Gem (and most other passive 
    seismic/acoustic data) are in UTC; time zones are not supported.
    """
    if type(nums) == int:
        nums = np.array([nums])
    if(len(station) == 0):
        station = SN
    fnList = _find_nonmissing_files(path, SN, nums)

    ## at this point, if we don't have any files, raise a missing file exception
    if len(fnList) == 0:
        raise MissingRawFiles(str(path) + ': ' + str(nums))
    while True:
        if len(fnList) == 0: # at this point, if we have no files, they're all corrupt. 
            raise CorruptRawFile(str(path) + ': ' + str(nums))
        try:
            version = _read_format_version(fnList[0])
            config = _read_config(fnList[0])
        except: # if we can't read the config for the first file here, drop it and try the next one
            fnList = fnList[1:] # 
        else:
            break
    if version in ['0.85C', '0.9', '0.91']:
        L = _read_several(fnList, require_gps = require_gps)# same function works for all
    elif (version == '0.85') | (version == '0.8') :
        L = _read_several(fnList, version = version, require_gps = require_gps) # same function works for both
    else:
        raise Exception(fnList[0] + ': Invalid or missing data format')

    ## stop early if we don't have data to process
    if len(L['data']) == 0:
        #return L
        raise CorruptRawFileNoGPS('No GPS information in data files ' + str(nums) + ' for SN "' + SN + '" in raw directory ' + str(path))

    L, timing_info = _assign_times(L)
    
    for tr in L['data']:
        ## populate the rest of the trace stats
        tr.stats.station = station
        tr.stats.location = location # this may well be ''
        tr.stats.network = network # can be '' for now and set later
    ## add bitweight and config info to header
    bitweight_info = get_bitweight_info(SN, config)
    header = L['header']
    for key in bitweight_info.keys():
        L['header'][key] = bitweight_info[key]
    for key in config.keys():
        L['header'][key] = config[key]
    L['header']['file_format_version'] = version

    ## done processing
    if return_debug_output:
        L['debug_output'] = timing_info
    return L

ReadGem = read_gem ## alias, v1.0.0
#################################################################

def _new_gem_var():
    tr = obspy.Trace()
    tr.stats.delta = 0.01
    gps = pd.DataFrame(columns=['year', 'date', 'lat', 'lon'])
    metadata = pd.DataFrame(columns=['millis', 'batt', 'temp', 'A2', 'A3', \
                                     'maxWriteTime', 'minFifoFree', 'maxFifoUsed', \
                                     'maxOverruns', 'gpsOnFlag', 'unusedStack1',\
                                     'unusedStackIdle', 't'])
    output = {'data': tr,
              'metadata': metadata,
              'gps': gps
    }
    return output


def _read_SN(fn):
    try:
        SN_line = pd.read_csv(fn, delimiter = ',', skiprows = 4, nrows=1, dtype = 'str', names=['s', 'SN'], encoding_errors='ignore')
        SN = SN_line['SN'][0]
    except:
        raise CorruptRawFile(fn + ': missing serial number')
    return SN

def _read_format_version(fn):
    #"""
    #0.91: like 0.9, but millisecond counts for GPS lines are floats
    #0.9: like 0.85C, but includes C line (possible that C line appeared in some earlier 0.85C files)
    #0.85C: data format is more compact than 0.85; otherwise like 0.85
    #0.85: ser. num. as extension, added A2 and A3 to metadata, otherwise same as 0.8
    #0.8: file extension .TXT, 1-hour files
    #"""
    versionLine = pd.read_csv(fn, delimiter = ',', nrows=1, dtype = 'str', names=['s'], encoding_errors='ignore')
    version = versionLine['s'][0][7:]
    return version
    
def _read_config(fn):
    config = pd.Series({'gps_mode': 1,
                    'gps_cycle' : 15,
                    'gps_quota' : 20,
                    'adc_range' : 0,
                    'led_shutoff' : 0,
                    'serial_output' : 0}) ## default config: it's fairly safe to use this as the default because any other configuration would require ...
    for j in range(10):
        line = pd.read_csv(fn, skiprows = j+1, nrows=1, delimiter = ',', dtype = 'str', names = ['na', 'gps_mode','gps_cycle','gps_quota','adc_range','led_shutoff','serial_output'], encoding_errors='ignore')
        if line.iloc[0,0] == 'C':
            #config = line.iloc[0,1:]
            config = {key:int(line[key]) for key in list(line.keys())[1:]}
            break
    return config


def _find_nonmissing_files(path, SN, nums):
    ## list all Gem files in the path
    #fnList = glob.glob(path + '/' + 'FILE[0-9][0-9][0-9][0-9].[0-9][0-9][0-9]')
    fnList = glob.glob(path + '/' + 'FILE[0-9][0-9][0-9][0-9].???')
    fnList.sort()
    fnList = np.array(fnList)
    
    ## find out what all the files' SNs are
    ext = np.array([x[-3:] for x in fnList])
    for i in range(len(ext)):
        if ext[i] == 'TXT':
            try:
                ext[i] = _read_SN(fnList[i])
            except: # if we're here, it's a corrupt/empty file
                pass # do nothing--it won't make it into goodFnList
            ## check the files for SN and num
    goodFnList = []
    for i, fn in enumerate(fnList):
        fnNum = int(fn[-8:-4])
        fnSN = ext[i]
        if (fnNum in nums) & (fnSN == SN):
            goodFnList.append(fn)
    if len(goodFnList) == 0:
        print('No good data files found for specified nums and SN ' + SN)
        return []
        ## fix this to be an exception or warning?
    ## make sure they aren't empty
    goodNonemptyFnList = []
    for fn in goodFnList:
        if os.path.getsize(fn) > 10: # to be safe, anything under 10 bytes is treated as empty 
            goodNonemptyFnList.append(fn)
    if(len(goodNonemptyFnList) == 0):
        ## warning
        print('No non-empty files')
        return []
    if(len(goodNonemptyFnList) < len(goodFnList)):
        print('Some files are empty, skipping')
    return goodNonemptyFnList    



def _unwrap_millis(new, old, maxNegative = 2**12, rollover = 2**13):
    ## maxNegative is the greatest allowable negative difference.
    ## negative differences can happen between data and GPS lines, or between metadata and data lines.
    return old + ((new - (old % rollover) + maxNegative) % rollover) - maxNegative

def _check_gps(line): # return True if GPS line is good
    #G,msPPS,msLag,yr,mo,day,hr,min,sec,lat,lon
    return not ((line[8] == 0) or (line[8] > 90) or (line[8] < -90) or # lat
                (line[9] == 0) or (line[9] > 180) or (line[9] < -180) or # lon
                (line[1] > 1000) or (line[1] < 0) or # lag
                (line[2] > 2040) or (line[2] < 2014) or # year
                (line[3] > 12) or (line[3] < 1) or # month
                (line[4] > 31) or (line[4] < 1) or # day
                (line[5] > 24) or (line[5] < 0) or # hour
                (line[6] > 60) or (line[6] < 0) or # minute
                (line[7]>60) or (line[7]<0) or (line[7]!=np.round(line[7]))) # second


def _make_gps_time(line):
    try:
        return obspy.UTCDateTime(int(line[2]), int(line[3]), int(line[4]), int(line[5]), int(line[6]), int(line[7]))
    except:
        return np.NaN

def _read_with_cython(filename, require_gps = True):
    """
    Read a Gem logfile.

    Parameters
    ----------
    filename : str or pathlib.Path
        Filepath of a file containing data to read.
    offset : int, default 0
        A timing offset to include in the millisecond timestamp values.

    Returns
    -------
    dict of dataframes
        - data: the analog readings and associated timings
        - metadata: datalogger metadata
        - gps: GPS timing and location values
    """
    try:
        from gemlog.parsers import parse_gemfile
    except ImportError:
        raise ImportError(
            "gemlog's C-extensions are not available. Reinstall gemlog with "
            "C-extensions to use this function."
        )

    # use cythonized reader file instead of pd.read_csv and slow string ops
    values, types, millis = parse_gemfile(str(filename).encode('utf-8'))
    if values.shape[0] == 0:
        raise EmptyRawFile(filename)
    if (b'G' not in types) and require_gps:
        raise CorruptRawFileNoGPS(filename)
    df = pd.DataFrame(values, columns=range(2, 13))
    # note that linetype has type bytes here, not str like in the pandas func
    df['linetype'] = types
    df['millis-sawtooth'] = millis
    #return _process_gemlog_data(df, offset)
    return df


def _read_0_8_with_pandas(filename, require_gps = True):
    ## This procedure is different enough from read_with_pandas that they are not interchangeable.
    # skiprows is important so that the header doesn't force dtype=='object'
    # the C engine for pd.read_csv is fast but crashes sometimes. Use the python engine as a backup.
    try:
        df = pd.read_csv(filename, names=range(13), low_memory=False, skiprows=6, encoding_errors='ignore')
    except Exception:
        try:
            df = pd.read_csv(filename, names=range(13), engine='python', skiprows=6,
                             error_bad_lines = False, warn_bad_lines = False, encoding_errors='ignore')
        except:
            raise CorruptRawFile(filename)
    if df.shape[0] == 0:
        raise EmptyRawFile(filename)

    df = df.iloc[np.where(~np.isnan(df.iloc[:,1]))[0],:]
    df['linetype'] = df.iloc[:,0].copy()
    df = df.iloc[:,1:]
    
    if ('G' not in set(df.loc[:,'linetype'])) and require_gps:
        raise CorruptRawFileNoGPS(filename)
    
    df = df.loc[df.loc[:,'linetype'].isin(['D', 'M', 'G']), :]

    ## most of the runtime is before here
    # unroll the ms rollover sawtooth
    df['millis-sawtooth'] = df.iloc[:,0]
    df['millis-sawtooth'] = df['millis-sawtooth'].astype(int)
    return df

def _read_with_pandas(filename, require_gps = True):
    # skiprows is important so that the header doesn't force dtype=='object'
    # the C engine for pd.read_csv is fast but crashes sometimes. Use the python engine as a backup.
    try:
        df = pd.read_csv(filename, names=range(13), low_memory=False, skiprows=6, encoding_errors='ignore')
    except Exception:
        try:
            df = pd.read_csv(filename, names=range(13), engine='python', skiprows=6,
                             error_bad_lines = False, warn_bad_lines = False, encoding_errors='ignore')
        except:
            raise CorruptRawFile(filename)
    if df.shape[0] == 0:
        raise EmptyRawFile(filename)

    try:
        df['linetype'] = [value[0] for value in df[0]] # exception if any 'value' is not a string
    except:
        raise CorruptRawFile(filename)
    if ('G' not in set(df.loc[:,'linetype'])) and require_gps:
        raise CorruptRawFileNoGPS(filename)
    
    df = df.loc[df.loc[:,'linetype'].isin(['D', 'M', 'G']), :]
    ## most of the runtime is before here
    # unroll the ms rollover sawtooth
    df['millis-sawtooth'] = np.where(df['linetype'] == 'D',df[0].str[1:],df[1]).astype(int)
    return df

def _read_single(filename, offset=0, require_gps = True, version = '0.9'):
    """
    Read a Gem logfile.

    Parameters
    ----------
    filename : str or pathlib.Path
        Filepath of a file containing data to read.

    offset : int, default 0
        A timing offset to include in the millisecond timestamp values.

    version: str, default '0.9'
        What raw format version to read. The raw format is in the first line of the raw file.

    require_gps : bool, default True
        Indicator of whether gps tags are required for reading the file. If True and gps tags are
        missing, raise a CorruptRawFileNoGPS exception.
    
    Returns
    -------
    dict of dataframes

        - data: the analog readings and associated timings
        - metadata: datalogger metadata
        - gps: GPS timing and location values
    """
    # Try each of the three file readers in order of decreasing speed but
    # probably increasing likelihood of success.

    if version in ['0.91', '0.9', '0.85C']:
        readers = [ _read_with_cython, _read_with_pandas]#, _slow__read_single_v0_9 ]
    else:
        readers = [_read_0_8_with_pandas]

    for reader in readers:
        try:
            df = reader(filename, require_gps)
            output = _process_gemlog_data(df, offset, version = version, require_gps = require_gps)
        except (EmptyRawFile, FileNotFoundError, CorruptRawFileNoGPS, KeyboardInterrupt):
            # If the file is definitely not going to work, exit early and
            # re-raise the exception that caused the problem
            raise
        except Exception:
            pass
        else: # if we're here, the file read worked. it may be invalid though.
            if (len(output['gps'].lat) == 0) and require_gps:
                raise CorruptRawFile(filename)

            # implement a small timing correction (depends on the raw format version)
            output['data'][:,0] += _time_corrections[version] 
            return output


    raise CorruptRawFile(filename)

def _process_gemlog_data(df, offset=0, version = '0.9', require_gps = True):
    ## figure out what settings to used according to the raw file format version
    if version in ['0.9', '0.85C']:
        rollover = 2**13
        M_cols = ['millis', 'batt', 'temp', 'A2', 'A3',
                  'maxWriteTime', 'minFifoFree', 'maxFifoUsed',
                  'maxOverruns', 'gpsOnFlag', 'unusedStack1', 'unusedStackIdle']
    elif version == '0.85':
        rollover = 2**16
        M_cols = ['millis', 'batt', 'temp', 'A2', 'A3',
                  'maxWriteTime', 'minFifoFree', 'maxFifoUsed',
                  'maxOverruns', 'gpsOnFlag', 'unusedStack1', 'unusedStackIdle']
    elif version == '0.8':
        rollover = 2**16
        M_cols = ['millis', 'batt', 'temp', 'maxWriteTime', 'minFifoFree', 'maxFifoUsed',
                  'maxOverruns', 'gpsOnFlag', 'unusedStack1', 'unusedStackIdle']
    else:
        raise CorruptRawFile('Invalid raw format version')
        
    # unroll the ms rollover sawtooth
    df['millis-stairstep'] = (df['millis-sawtooth'].diff() < -(rollover/2)).cumsum()
    df['millis-stairstep'] -= (df['millis-sawtooth'].diff() > (rollover/2)).cumsum()
    df['millis-stairstep'] *= rollover
    df['millis-corrected'] = df['millis-stairstep'] + df['millis-sawtooth']
    first_millis = df['millis-corrected'].iloc[0]
    df['millis-corrected'] += (
        (offset-first_millis)
        + ((first_millis-(offset % rollover)+rollover/2) % rollover)
        - rollover/2
    )
    # groupby is somewhat faster than repeated subsetting like
    # df.loc[df['linetype'] == 'D', :], ...
    grouper = df.groupby('linetype')
    # the python-based reader uses strings for linetype but the cython version
    # uses bytes.  figure out which one we need.
    # Wonder if this could be cleaned up by coercing it to str.
    if isinstance(df['linetype'].iloc[0], bytes):
        Dkey = b'D'
        Gkey = b'G'
        Mkey = b'M'
        data_col = 2
    else:
        Dkey, Gkey, Mkey = 'DGM'
        data_col = 1
    if version in ['0.8', '0.85']:
        data_col = 2
    D = grouper.get_group(Dkey)
    M = grouper.get_group(Mkey)
    #_breakpoint()
    # pick out columns of interest and rename
    D_cols = ['msSamp', 'ADC']

    # column names are currently integers (except for the calculated cols)
    D = D[['millis-corrected', data_col]]
    D.columns = D_cols
    M = M[['millis-corrected'] + list(range(2, len(M_cols)+1))]
    M.columns = M_cols

    # now that columns aren't mixed dtype anymore, convert to numeric where possible
    D = D.apply(pd.to_numeric)
    M = M.apply(pd.to_numeric)


    # process data (version-dependent)
    if version in ['0.9', '0.85C']: # don't integrate the data if version is 0.8, 0.85
        D['ADC'] = D['ADC'].astype(float).cumsum()

    ## gps stuff
    G_cols = ['msPPS', 'msLag', 'year', 'month', 'day', 'hour', 'minute', 'second', 'lat', 'lon']
    def make_gps_time(row):
        try:
            return obspy.UTCDateTime(*row)
        except Exception:
            return np.NaN
    try:
        G = grouper.get_group(Gkey)
        G = G[['millis-corrected'] + list(range(2, len(G_cols)+1))]
        G.columns = G_cols
        G = G.apply(pd.to_numeric)
        # filter bad GPS data and combine into datetimes
        valid_gps = _valid_gps(G)
        G = G.loc[valid_gps, :]
        G['t'] = G.iloc[:, 2:8].astype(int).apply(make_gps_time, axis=1)
        G = G.reset_index().astype('float')
    except:
        if require_gps:
            raise CorruptRawFileNoGPS()
        else:
            G = pd.DataFrame(columns = G_cols)
        
    return {'data': np.array(D), 'metadata': M.reset_index().astype('float'), 'gps': G}

def _valid_gps(G):
    # vectorized GPS data validation
    # basic lower and upper bounds:
    limits = {
        'lat': (-90, 90),
        'lon': (-180, 180),
        'msLag': (0, 1000),
        'year': (2014, 2040),
        'month': (1, 12),
        'day': (1, 31),
        'hour': (0, 24),
        'minute': (0, 60),
        'second': (0, 60),
    }
    bad_gps = False
    for key, (lo, hi) in limits.items():
        data = G[key]
        bad_gps |= (data < lo) | (data > hi)

    # custom limits:
    bad_gps |= (
        (G['lat'] == 0) |
        (G['lon'] == 0) |
        (G['second'] != np.round(G['second']))
    )
    return ~bad_gps


def _slow__read_single_v0_9(filename, offset=0, require_gps = True):
    ## this should only be used as a reference
    ## pre-allocate the arrays (more space than is needed)
    M = np.ndarray([15000,12]) # no more than 14400
    G = np.ndarray([15000,11]) # no more than 14400
    D = np.ndarray([750000,2]) # expected number 7.2e5
    d_index = 0
    m_index = 0
    g_index = 0
    millis = offset % 2**13 # 2020-11-05
    ## open the file for reading
    with open(filename, 'r', newline = '', encoding='ascii', errors = 'ignore') as csvfile:
        lines = csv.reader(csvfile, delimiter = ',', quoting = csv.QUOTE_NONE)
        i = 0
        for line in lines:
            ## determine the line type, and skip if it's not necessary data (e.g. debugging info)
            try:
                lineType = line[0][0] # if this fails, it means the line is empty or at least invalid
            except:
                continue
            if not (lineType in ['D', 'G', 'M']):
                continue
            ## remove the line type ID and make into a nice array
            if lineType == 'D':
                line[0] = line[0][1:]
            else:
                line = line[1:]
            line = np.array([float(x) for x in line])
            ## unwrap the millis count (always first element of line)
            millis = _unwrap_millis(line[0], millis)
            line[0] = millis
            ## write the line to its matrix
            if lineType == 'D':
                D[d_index,:] = line
                d_index += 1
            elif lineType == 'M':
                M[m_index,:] = line
                m_index += 1
            elif (lineType == 'G') and _check_gps(line):
                G[g_index,:10] = line
                G[g_index,10] = _make_gps_time(line)
                g_index += 1
    if d_index == 0:
        raise EmptyRawFile(filename)
    if (g_index == 0) and require_gps:
        raise CorruptRawFileNoGPS(filename)
    #pdb.set_trace()
    ## remove unused space in pre-allocated arrays
    D = D[:d_index,:]
    G = pd.DataFrame(G[:g_index,:], columns = ['msPPS', 'msLag', 'year', 'month', 'day', 'hour', \
                                               'minute', 'second', 'lat', 'lon', 't'])
    M = pd.DataFrame(M[:m_index,:], columns = ['millis', 'batt', 'temp', 'A2', 'A3', \
                                               'maxWriteTime', 'minFifoFree', 'maxFifoUsed', \
                                               'maxOverruns', 'gpsOnFlag', 'unusedStack1',\
                                               'unusedStackIdle'])
    ## process data (version-dependent)
    D[:,1] = D[:,1].cumsum()
    return {'data': D, 'metadata': M, 'gps': G}

def _read_several(fnList, version = 0.9, require_gps = True):
    ## initialize the output variables
    D = np.ndarray([0,2]) # expected number 7.2e5
    header = _make_empty_header(fnList)
    G = _make_empty_gps()
    M = _make_empty_metadata()
    
    ## loop through the files
    startMillis = 0
    for i,fn in enumerate(fnList):
        print('File ' + str(i+1) + ' of ' + str(len(fnList)) + ': ' + fn)
        try:
            if str(version) in ['0.91', '0.9', '0.85C']:
                L = _read_single(fn, startMillis, require_gps = require_gps)
            elif str(version) in ['0.8', '0.85']:
                L = _read_single(fn, startMillis, require_gps = require_gps, version = version)
            else:
                raise CorruptRawFile('Invalid raw file format version: ' + str(version))
            if(L['data'][0,0] < startMillis):
                L['metadata'].millis += 2**13
                L['gps'].msPPS += 2**13
                L['data'][:,0] += 2**13
            dMillis = np.diff(L['data'][:,0])
            ## if the millis series is discontinuous, reject the file
            if any(dMillis < 0) or any(dMillis > 1000):
                raise CorruptRawFile(f'{fn} samples times are discontinuous')
            if len(L['gps'].t) == 0:
                raise CorruptRawFileNoGPS('No GPS data in ' + fn + ', skipping')

            ## make sure that the available GPS spans a significant fraction of the available
            ## data samples; otherwise the time interpolation will be unreliable
            ## the ratio threshold of 2 can be tuned. A threshold of 4 was too lax when tested on
            ## 2021 Fuego data (SN 179), resulting in artificial discontinuities between files of
            ## around 10-20 samples. A threshold less than 2 could make gemconvert refuse to convert
            ## a short recording period just under 30 minutes (stops recording just before the 
            ## second gps cycle)
            if (len(L['gps']) == 1) or (0.001024*(L['data'][-1,0] - L['data'][0,0]) / (L['gps'].t.iloc[-1] - L['gps'].t.iloc[0])) > 2:
                raise CorruptRawFileInadequateGPS(f'Insufficient GPS data in {fn}, skipping')
            try:
                ## run the GPS time vs millis regression
                reg, num_gps_nonoutliers, MAD_nonoutliers, resid, xx, yy = _robust_regress(L['gps'].msPPS, L['gps'].t)
            except:
                raise CorruptRawFileInadequateGPS('No useful GPS data in ' + fn + ', skipping')
            if ((0.001024*(L['data'][-1,0] - L['data'][0,0]) / (xx.iloc[-1] - xx.iloc[0])) > 2) \
               or (num_gps_nonoutliers < 10) \
               or MAD_nonoutliers > 0.01:
                raise CorruptRawFileInadequateGPS('No useful GPS data in ' + fn + ', skipping')
            header.loc[i, 'num_data_pts'] = L['data'].shape[0]
            header.loc[i, 'SN'] = _read_SN(fn)

        except KeyboardInterrupt:
            raise
        except CorruptRawFileInadequateGPS:
            print('Insufficient GPS data in ' + fn + ', skipping')
        except CorruptRawFileNoGPS:
            print('No GPS data in ' + fn + ', skipping')
        except:
            print('Failed to read ' + fn + ', skipping')
            _breakpoint()
        else:
            M = pd.concat((M, L['metadata']))
            G = pd.concat((G, L['gps']))
            D = np.vstack((D, L['data']))
            #_breakpoint()
            if L['gps'].shape[0] > 0:
                ## this can result in an error if there aren't enough good GPS data points
                startMillis = D[-1,0]
                header.loc[i, 'lat'] = np.median(L['gps']['lat'])
                header.loc[i, 'lon'] = np.median(L['gps']['lon'])
                header.loc[i, 'start_ms'] = L['data'][0,0] # save this as a millis first, then convert
                header.loc[i, 'end_ms'] = L['data'][-1,0]
                header.loc[i, 'drift_deg3'] = reg[0]
                header.loc[i, 'drift_deg2'] = reg[1]
                header.loc[i, 'drift_deg1'] = reg[2]
                header.loc[i, 'drift_deg0'] = reg[3]
                header.loc[i, 'drift_resid_std'] = np.std(resid)
                header.loc[i, 'drift_resid_MAD'] = np.max(np.abs(resid))
                header.loc[i, 'num_gps_pts'] = len(L['gps'].msPPS)
                header.loc[i, 'drift_resid_MAD_nonoutliers'] = MAD_nonoutliers
                header.loc[i, 'num_gps_nonoutliers'] = num_gps_nonoutliers
        ## end of fn loop
    #_breakpoint()
    return {'metadata':M, 'gps':G, 'data': D, 'header': header}

def _robust_regress(x, y, z = 4, recursive_depth = np.inf, verbose = False):
    # goal: a quadratic regression that is robust to RARE outliers, especially for GPS data
    # scipy.stats.theilslopes (median-based) looks problematic because the median is only affected
    # by the central data point and doesn't benefit from the other samples' information. Also, GPS
    # data slopes are weirdly distributed.
    # In this function, z is the z-score (number of standard deviations) for defining outliers.

    ### z < 3 has an off-chance of repeated trimming with few data points remaining! don't do that.

    ## Calculate regression line and residuals.
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        reg = np.polynomial.polynomial.polyfit(x, y, 3)[::-1]
        #linreg = scipy.stats.linregress(x, y)
        #resid = y - (linreg.intercept + x * linreg.slope)
        #reg = np.polyfit(x, y, 3)
    resid = y - (reg[3] + x * reg[2] + x**2 * reg[1] + x**3 * reg[0])
    #resid = y - _apply_fit(x, reg) # doesn't work because reg is not dict

    ## If any are found to be outliers, remove them and recalculate recursively.
    outliers = np.abs(resid) > (z*np.std(resid))
    if any(outliers) and (recursive_depth > 0):
        if verbose:
            print(f'depth {recursive_depth}, num_outliers {np.sum(outliers)}')
        return _robust_regress(x[~outliers], y[~outliers], z, recursive_depth = recursive_depth - 1, verbose = verbose)
    else:
        return (reg, len(x), np.max(np.abs(resid)), resid, x, y)

def _no_drift(x):
    return x * 0.001024

def _plot_drift(file_list, z = 4, recursive_depth = np.inf):
    if type(file_list) is str:
        file_list = [file_list]
    output = _read_several(file_list)
    g = output['gps']
    xg = np.array(g['msPPS'])
    yg = np.array(g['t']) # GPS time
    xd = output['data'][:,0]

    plt.subplot(2,1,1)
    plt.plot(xd, _apply_segments(xd, output['header']) - _no_drift(xd) - yg[0])
    plt.plot(xg, _apply_segments(xg, output['header']) - _no_drift(xg) - yg[0], 'k.')

    drift_gps = _apply_segments(xg, output['header']) - _no_drift(xg) - yg[0]
    drift_slope = (drift_gps[-1] - drift_gps[0]) / (xg[-1] - xg[0])
    plt.subplot(2,1,2)
    plt.plot(xd, _apply_segments(xd, output['header']) - _no_drift(xd) - yg[0] - xd * drift_slope)
    plt.plot(xg, _apply_segments(xg, output['header']) - _no_drift(xg) - yg[0] - xg * drift_slope, 'k.')

def _apply_fit(x, model):
    return model['drift_deg0'] + model['drift_deg1'] * x + model['drift_deg2'] * x**2 + model['drift_deg3'] * x**3
             
def _apply_segments(x, model):
    y = np.zeros(len(x))
    y[:] = np.nan
    for i in range(len(model['start_ms'])):
        w = (x >= model['start_ms'][i]) & (x <= model['end_ms'][i])
        #y[w] = model['drift_deg0'][i] + model['drift_deg1'][i] * x[w] + model['drift_deg2'][i] * x[w]**2 + model['drift_deg3'][i] * x[w]**3
        y[w] = _apply_fit(x[w], model.iloc[i,:])
    return y
    
def _assign_times(L):
    #_breakpoint()
    fnList = L['header'].file
    if L['gps'].shape[0] == 0:
        raise Exception('No GPS data in files ' + fnList[0] + '-' + fnList[-1] + '; stopping conversion')
    
    G = _reformat_GPS(L['gps'])
    try:
        breaks = _find_breaks_(L)
    except:
        raise CorruptRawFile('Problem between ' + fnList[0] + '-' + fnList[-1] + '; stopping before this interval. Break between recording periods? Corrupt files?')
    piecewiseTimeFit = L['header']
    L['metadata']['t'] = _apply_segments(L['metadata']['millis'], piecewiseTimeFit)
    header = L['header']
    header['t1'] = _apply_segments(header.start_ms, piecewiseTimeFit)
    header['t2'] = _apply_segments(header.end_ms, piecewiseTimeFit)
    L['header'] = header
    
    ## Interpolate data to equal spacing to make obspy trace.
    ## Note that data gaps just get interpolated through as a straight line. Not ideal.
    D = L['data']
    D = np.hstack((D, _apply_segments(D[:,0], piecewiseTimeFit).reshape([D.shape[0],1])))
    timing_info = [L['gps'], L['data'], breaks, piecewiseTimeFit]
    L['data'] = _interp_time(D) # returns stream, populates known fields: channel, delta, and starttime
    L['gps'] = G
    return (L, timing_info)
    


#########################################################
def _interp_time(data, t1 = -np.Inf, t2 = np.Inf, min_step = 0, max_step = 0.025):
    ## min_step, max_step are min/max interval allowed before a break is identified
    eps = 0.001 # this might need some adjusting to prevent short data gaps
    ## break up the data into continuous chunks, then round off the starts to the appropriate unit
    ## t1 is the first output sample; should be first integer second after or including the first sample (ceiling)
    w_nonnan = ~np.isnan(data[:,2])
    t_in = data[w_nonnan,2]
    p_in = data[w_nonnan,1]
    _breakpoint()
    #t1 = np.trunc(t_in[t_in >= t1][0]+1-0.01) ## added on 2019-09-11. 2021-07-11: this line is responsible for losing samples before the start of a second
    t1 = t_in[t_in >= (t1 - 0.01 - eps)][0]
    t2 = t_in[t_in <= (t2 + .01 + eps)][-1] # add a sample because t2 is 1 sample before the hour
    breaks_raw = np.where((np.diff(t_in) > max_step) | (np.diff(t_in) < min_step) )[0] # 2020-11-04: check for backwards steps too
    breaks = breaks_raw[(t_in[breaks_raw] > t1) & (t_in[breaks_raw+1] < t2)]
    starts = np.hstack([t1, t_in[breaks+1]]) # start times of continuous chunks
    ends = np.hstack([t_in[breaks], t2]) # end times of continuous chunks
    w_same = (starts != ends)
    starts = starts[w_same]
    ends = ends[w_same]
    ## make an output time vector excluding data gaps, rounded to the nearest samples
    #t_interp = np.zeros(0)
    output = obspy.Stream()
    for i in range(len(starts)):
        w = (t_in >= (starts[i] - 0.01 - eps)) & (t_in <= (ends[i] + 0.01 - eps))
        try:
            f = scipy.interpolate.CubicSpline(t_in[w], p_in[w])
        except:
            _breakpoint()
            continue
            ##if not _debug: # so pdb doesn't end immediately with this exception
            ##    raise(Exception('_interp_time failed between ' +str(obspy.UTCDateTime(starts[i])) +\
            ##                    ' and ' + str(obspy.UTCDateTime(ends[i]))))
        #t_interp = np.arange(starts[i], ends[i] + eps, 0.01) # this is a bug in np.arange--intervals can be inconsistent when delta is float. This can result in significant timing errors, especially for long traces.
        t_interp = np.round(starts[i] + np.arange(np.trunc((ends-starts)[i]/0.01)) * 0.01, 2)
        p_interp = np.array(f(t_interp).round(), dtype = 'int32')
        tr = obspy.Trace(p_interp)
        tr.stats.starttime = t_interp[0]
        tr.stats.delta = 0.01
        tr.stats.channel = 'HDF'
        output += tr
    return output
## old code in and below the for loop. no good because a trace is only good for a continuous block. We may have a break, so we need a stream.
#        t_interp = np.concatenate([t_interp, t_new])
#    t_interp = t_interp[(t_interp >= (t1-eps)) & (t_interp < (t2 + eps))]
#    ## interpolate to find pressure at these sample times
#    f = scipy.interpolate.CubicSpline(t_in, p_in)
#    p_interp = np.array(f(t_interp).round(), dtype = 'int32')
#    tr = obspy.Trace(p_interp)
#    tr.stats.starttime = t_interp[0]
#    tr.stats.delta = 0.01
#    tr.stats.channel = 'HDF'
#    return tr

###############################################################
#  version bitweight_Pa  bitweight_V min_SN max_SN
#1    0.50  0.003256538 7.362894e-08      3      7
#2    0.70  0.003256538 7.362894e-08      8     14
#3    0.80  0.003256538 7.362894e-08     15     19
#4    0.82  0.003256538 7.362894e-08     20     37
#5    0.90  0.003543324 7.362894e-08     38     40
#6    0.91  0.003543324 7.362894e-08     41     43
#7    0.92  0.003543324 7.362894e-08     44     46
#8    0.98  0.003501200 7.275362e-08     47     49
#9    0.99  0.003501200 7.275362e-08     50    54
#10    0.991  0.003501200 7.275362e-08     52    54
#11    0.992  0.003501200 7.275362e-08     55    57
#12    1.00  0.003501200 7.275362e-08     58    107
#13    1.00  0.003501200 7.275362e-08     108    Inf
#'bitweight_Pa': [0.003256538, 0.003256538, 0.003256538, 0.003256538, 0.003543324, 0.003543324, 0.003543324, 0.003501200, 0.003501200, 0.003501200, 0.003501200, 0.003501200, 0.003501200],

#bitweight_V = [7.362894e-08, 7.362894e-08, 7.362894e-08, 7.362894e-08, 7.362894e-08, 7.362894e-08, 7.362894e-08, 7.275362e-08, 7.275362e-08, 7.275362e-08, 7.275362e-08,7.275362e-08,7.275362e-08]
__AVCC__ = np.array([3.373, 3.373,  3.373,  3.373,  3.373,  3.373,  3.373, 7.275362e-08, 7.275362e-08, 7.275362e-08, 7.275362e-08, 7.275362e-08, 7.275362e-08]),
def __AVCC__(version):
    if version >= 0.90:
        return 3.1
    else:
        return 3.373

def __gain__(version):
    if version >= 0.98:
        return 1 + 50/0.470
    else:
        return 1 + 49.4/0.470

def get_gem_specs(SN):
    """
    Retrieve specs for a given Gem serial number. 
    
    Parameters
    ----------
    SN : str or int
    
    Returns
    -------
    dict with the following elements:

        - version : Gem version number
        - bitweight_V : voltage resolution for this Gem version [Volts per count]
        - bitweight_Pa : pressure resolution for this Gem version [Pascals per count]
    """
    versionTable = {'version': np.array([0.5, 0.7, 0.8, 0.82, 0.9, 0.91, 0.92, 0.98, 0.99, 0.991, 0.992, 1, 1.01]),
                    'min_SN': np.array([3, 8, 15, 20, 38, 41, 44, 47, 50, 52, 55, 58, 108]),
                    'max_SN': np.array([7, 14, 19, 37, 40, 43, 46, 49, 51, 54, 57, 107, np.Inf])
    }
    version = versionTable['version'][(int(SN) >= versionTable['min_SN']) & (int(SN) <= versionTable['max_SN'])][0]
    bitweight_V = 0.256/2**15/__gain__(version)
    sensitivity = __AVCC__(version)/7.0 * 45.13e-6 # 45.13 uV/Pa is with 7V reference from Marcillo et al., 2012
    return { 'version': version,
             'bitweight_V': bitweight_V,
             'bitweight_Pa': bitweight_V/sensitivity
    }


## Eventually, get_bitweight_info should perform all the functions of the R gemlog equivalent...but not yet.
def get_bitweight_info(SN, config, units = 'Pa'):
    if config['adc_range'] == 0: # high gain
        multiplier = 1
    elif config['adc_range'] == 1: # low gain
        multiplier = 2
    else:
        #pdb.set_trace()
        raise BaseException('Invalid Configuration')
    specs = get_gem_specs(SN)
    specs['bitweight_Pa'] *= multiplier
    specs['bitweight_V'] *= multiplier
    if units.upper() == 'PA':
        specs['bitweight'] = specs['bitweight_Pa']
    elif units.upper() == 'V':
        specs['bitweight'] = specs['bitweight_V']
    elif units.lower() == 'counts':
        specs['bitweight'] = 1
    else:
        raise BaseException('Invalid Units')
    return specs

def _reformat_GPS(G_in):
    t = [obspy.UTCDateTime(tt) for tt in G_in['t']]
    date = [tt.julday + tt.hour/24.0 + tt.minute/1440.0 + tt.second/86400.0 for tt in t]
    G_dict = {'year': np.array([int(year) for year in G_in.year]),
              'date': np.array(date),
              'lat': np.array(G_in.lat),
              'lon': np.array(G_in.lon),
              't': np.array(t)}
    return pd.DataFrame.from_dict(G_dict)

def _find_breaks_(L):
    _breakpoint()
    ## breaks are specified as their millis for comparison between GPS and data
    ## sanity check: exclude suspect GPS tags
    t = np.array([obspy.UTCDateTime(tt) for tt in L['gps'].t])
    tPad = np.array([t[0]] + list(t) + [t[-1]])
    try:
        badTags = ((t > obspy.UTCDateTime.now()) | # no future dates 
                   (L['gps'].year < 2015) | # no years before the Gem existed
                   ((np.abs(t - tPad[:-2]) > 86400) & (np.abs(t - tPad[2:]) > 86400)) | # exclude outliers
                   (L['gps'].lat == 0) | # exclude points within ~1m of the equator
                   (L['gps'].lon == 0)) # exclude points within ~1m of the prime meridian
        L['gps'] = L['gps'].iloc[np.where(~badTags)[0],:]
    except:
        _breakpoint()
    mD = np.array(L['data'][:,0]) # data millis
    dmD = np.diff(mD)
    starts = np.array([])
    ends = np.array([])
    dataBreaks = np.where((dmD > 25) | (dmD < 0))[0]
    if 0 in dataBreaks:
        mD = mD[1:]
        dmD = dmD[1:]
        dataBreaks = dataBreaks[dataBreaks != 0] - 1
    if (len(dmD) - 1) in dataBreaks:
        mD = mD[:-1]
        dmD = dmD[:-1]
        dataBreaks = dataBreaks[dataBreaks != len(dmD)]
    #_breakpoint()
    for i in dataBreaks:
        starts = np.append(starts, np.max(mD[(i-1):(i+2)]))
        ends = np.append(ends, np.min(mD[(i-1):(i+2)]))
    tG = np.array(L['gps'].t).astype('float') # gps times
    mG = np.array(L['gps'].msPPS).astype('float') # gps millis
    dmG_dtG = np.diff(mG)/np.diff(tG) * 1.024 # correction for custom millis in gem firmware (1024 us/ms)
    gpsBreaks = np.argwhere(np.isnan(dmG_dtG) | # missing data...unlikely
                            ((np.diff(tG) > 50) & ((dmG_dtG > 1000.1) | (dmG_dtG < 999.9))) | # 100 ppm drift between cycles
                            ((np.diff(tG) <= 50) & ((dmG_dtG > 1002) | (dmG_dtG < 998))) # most likely: jumps within a cycle (possibly due to leap second)
    )
    min_possible_start = mD.min() # this only changes if a GPS break around the first GPS sample invalidates preceding D samples
    for i in gpsBreaks:
        i = int(i)
        ## This part is tricky: what if a gpsEnd happens between a dataEnd and dataStart?
        ## Let's be conservative: if either the gpsEnd or gpsStart is within a bad data interval, or what if they bracket a bad data interval?
        ## choose them so that they exclude the most data
        #breakpoint()
        overlaps = (mG[i] <= starts) & (mG[i+1] >= ends)
        if(np.any(overlaps)):
            w = np.argwhere(overlaps)
            starts[w] = max(np.append(starts[w], mG[(i-1):(i+2)].max()))
            ends[w] = max(np.append(ends[w], mG[(i-1):(i+2)].min()))
        else:
            wmin = np.argwhere(tG > tG[i])
            ## If a file's very last gps fix triggered a break, an exception would be raised here.
            ## This is unlikely and it's not clear now what the right thing to do is. So not implemented.
            try:
                starts = np.append(starts, mG[wmin][tG[wmin] == tG[wmin].min()][-1]) # [-1] just in case tG values are repeated
            except:
                _breakpoint()
            wmax = np.argwhere(tG < tG[i+1])
            ## It's possible for a gps break to occur so early that no good fixes occur before the
            ## break (e.g., leap second change). In that case, pre-break data are unrecoverable.
            if len(wmax) == 0: # empty wmax means no fixes before break
                min_possible_start = tG[i+1]
            else: # normal: add an end at this break
                ends = np.append(ends, mG[wmax][tG[wmax] == tG[wmax].max()][-1]) # [-1] in case tG values are repeated
    starts = np.append(min_possible_start, starts)
    ends = np.append(ends, mD.max())
    return {'starts':starts, 'ends':ends}

def _make_empty_header(fnList):
    num_filler = np.zeros(len(fnList))
    return pd.DataFrame.from_dict({'file': fnList,
                                   'SN':['' for fn in fnList],
                                   'lat': num_filler,
                                   'lon': num_filler,
                                   'start_ms': num_filler,
                                   'end_ms': num_filler,
                                   #'drift_slope': num_filler,
                                   #'drift_intercept': num_filler,
                                   #'drift_slope_stderr': num_filler,
                                   'drift_deg0': num_filler,
                                   'drift_deg1': num_filler,
                                   'drift_deg2': num_filler,
                                   'drift_deg3': num_filler,
                                   'drift_resid_std': num_filler,
                                   'drift_resid_MAD': num_filler,
                                   'num_gps_pts': num_filler,
                                   'num_data_pts': num_filler
    })
def _make_empty_gps():
    return pd.DataFrame(columns = ['msPPS', 'msLag', 'year', 'month', 'day', 'hour', 'minute', \
                                   'second', 'lat', 'lon', 't'])
def _make_empty_metadata():
    return pd.DataFrame(columns = ['millis', 'batt', 'temp', 'A2', 'A3', 'maxWriteTime', \
                                   'minFifoFree', 'maxFifoUsed', 'maxOverruns', 'gpsOnFlag', \
                                   'unusedStack1', 'unusedStackIdle'])

## time corrections spec taken from 2021-07-07 PPS test.
_time_corrections = { # milliseconds
    '0.8':8.93,
    '0.85':8.93,
    '0.85C':8.93,
    '0.9':8.93,
    '0.91':8.93
}
    



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import obspy, glob, gemlog, os
import scipy.integrate
import scipy.interpolate
#from gemlog import *
def PlotAmp(DB):
    allSta = DB.station.unique()
    allSta.sort()
    for sta in DB.station.unique():
        w = np.where(DB.station == sta)[0]
        w.sort()
        plt.plot(DB.t1.iloc[w], np.log10(DB.amp_HP.iloc[w]), '.')
        print(str(sta) + ' ' + str(np.quantile(DB.amp_HP.iloc[w], 0.25)))
    plt.legend(allSta)
    plt.show()

import obspy
def CheckDiscontinuity(files):
    st = obspy.Stream()
    for fn in files:
        st += obspy.read(fn)
    st.merge()
    st = st.split()
    if len(st) > 1:
        for tr in st:
            print(tr.id + ' ' + tr.stats.starttime.isoformat() + '--' + tr.stats.endtime.isoformat())


def check_lags(DB, winlength = 1000, fl = 0.5, fh = 20, maxshift = 10, verbose = False):
    stations = DB.station.unique()
    nsta = len(stations)
    from obspy.signal.cross_correlation import xcorr, correlate, xcorr_max
    st = obspy.Stream()
    st.filter('bandpass', freqmin=fl, freqmax = fh)
    for fn in DB.filename:
        st += obspy.read(fn)
    t0 = min(DB.t1).replace(minute=0, second=0, microsecond=0)
    t = []
    lag = np.zeros([nsta, 1])
    xc_coef = np.zeros([nsta, 1])
    consistency = []
    N = int((max(DB.t2) - t0)/winlength)
    count = 0
    t1 = t0 + 1e-6
    #while t1 < (t0 + 10*winlength):#max(DB.t2):
    while t1 < max(DB.t2):
        count += 1
        t1 += winlength
        if verbose: print(str(count) + ' of ' + str(N))
        try:
            test_lags = []
            test_xc_coefs = []
            #st.trim(t1)
            st_test = st.slice(t1, t1 + winlength)
            for i in range(nsta):
                xc_output = xcorr_max(correlate(st_test.traces[i], st_test.traces[(i+1) % nsta], maxshift))
                test_lags.append(xc_output[0])
                test_xc_coefs.append(xc_output[1])
            t.append(t1)
            lag = np.hstack([lag, np.array(test_lags).reshape(nsta, 1)])
            xc_coef = np.hstack([xc_coef, np.array(test_xc_coefs).reshape(nsta, 1)])
            consistency.append(np.sum(test_lags))
        except:
            pass
    return([t, lag[:,1:], xc_coef[:,1:], np.array(consistency)])


def plot_lags(lag0, lag1, use_consistency = True):
    import matplotlib.pyplot as plt
    N=3
    ## Plot lags for the two conversions
    plt.subplot(2,2,1)
    ## consistency multiplier: 1 if consistent, NaN otherwise
    if use_consistency:
        c1 = (np.array(lag1[3])==0) / (np.array(lag1[3])==0)
        c0 = (np.array(lag0[3])==0) / (np.array(lag0[3])==0)
    else:
        c1 = 1
        c0 = 1
    for i in range(N):
        plt.plot(c1 * lag1[1][i,:]+i/6, '.')
    
    plt.subplot(2,2,3)
    for i in range(N):
        plt.plot(c0 * lag0[1][i,:]+i/6, '.')

    ## Plot the correlation coefficient
    plt.subplot(2,2,2)
    for i in range(N):
        plt.plot(c1*lag1[2][i,:], '.')

    plt.subplot(2,2,4)
    for i in range(N):
        plt.plot(c0 * lag0[2][i,:], '.')

    plt.legend([str(i) for i in range(N)], loc = 'lower right')
    plt.show()

def make_db(path, pattern = '*', savefile = None, verbose = False):
    """Create a database summarizing a set of converted data files.

    Parameters
    ----------
    path : str
        Path to folder containing converted data files to summarize.
    
    pattern : str, default '*'
        Glob-type pattern for selecting converted data files to summarize

    savefile : str, default None
        File name where database is written. Use 'savefile = None' to not save an output file.

    verbose : bool, default False
        If True, print progress updates.

    Returns
    -------
    pandas.DataFrame containing converted file database.
    """
    #path = 'mseed'
    #pattern = '*'
    files = glob.glob(path + '/' + pattern)
    files.sort()
    DB = []
    count = 0
    for file in files:
        try:
            tr = obspy.read(file)[0]
        except:
            print('failed to read ' + file + ', skipping')
            continue
        maxVal = tr.data.max()
        minVal = tr.data.min()
        tr.detrend('linear')
        tr.filter('highpass', freq=0.5)
        amp_HP = tr.std()
        row = pd.DataFrame([[file, tr.stats.station, tr.stats.location, amp_HP, maxVal, minVal, tr.stats.starttime, tr.stats.endtime]], columns = ['filename', 'station', 'location', 'amp_HP', 'max', 'min', 't1', 't2'])
        DB.append(row)
        if((count % 100) == 0 and verbose):
            print(str(count) + ' of ' + str(len(files)))
        count = count + 1
    DB = pd.concat(DB)
    if savefile is not None:
        DB.to_csv(savefile)
    return(DB)

################################################

################

def calc_channel_stats(DB, t1, t2):
    """
    Calculate uptime and other statistics for all channels in a database.

    Parameters
    ----------
    DB : pandas.DataFrame
        Output of make_db().
    
    t1 : time-like 
        Start time for which statistics should be calculated.

    t2 : time-like 
        End time for which statistics should be calculated.

    Returns
    -------
    pandas.DataFrame containing the following columns:

        - station : station name
        - goodData : proportion of time window (t1-t2) that is not obviously bad (e.g., clipped)
        - anyData : proportion of time window (t1-t2) for which data are available
        - q1 : first quartile amplitude 
        - q3 : third quartile amplitude
    """
    import obspy, glob
    import pandas as pd
    from obspy import UTCDateTime as T
    #t1 = '2020-04-14'
    #t2 = '2020-04-24T20:00:00'
    t1 = obspy.core.UTCDateTime(t1)
    t2 = obspy.core.UTCDateTime(t2)
    numHour = (t2 - t1)/3600.0
    DB.t1 = DB.t1.apply(T)
    DB.t2 = DB.t2.apply(T)
    DB.goodData = (DB.amp_HP > 0.5) & (DB.amp_HP < 2e4) & ((DB.t2 - DB.t1) > 3598) & ((DB.t2 - DB.t1) < 3602)
    DB.anyData = (DB.amp_HP > 0) 
    out = []
    for sta in DB.station.unique():
        w = np.where((DB.station == sta) & (DB.t1 > t1) & (DB.t2 < t2))[0]
        if(len(w) == 0):
            continue
        else:
            q1 = np.quantile(np.array(DB.amp_HP)[w], 0.25)
            q3 = np.quantile(np.array(DB.amp_HP)[w], 0.75)
            out.append(pd.DataFrame([[sta, np.sum(np.array(DB.goodData)[w])/numHour, np.sum(np.array(DB.anyData)[w])/numHour, q1, q3]], columns = ['station', 'goodData', 'anyData', 'q1', 'q3']))
    out = pd.concat(out)
    return(out)

_noise_path = os.path.join(os.path.dirname(__file__), "data", "noise")

def gem_noise(freq = None, spectype = 'power', version = '1.0', freq_min = None, freq_max = 50):
    """Calculate Gem self-noise as a one-sided spectrum, and integrated over 
    a defined frequency band

    Parameters
    ----------
    freq : numpy array
        Input frequencies for spectrum. If None, use default.
    spectype : str
        Type of spectrum and noise calculations: either 'power', 'amp', or 'dB' 
    version : str
        Gem version to find noise spectrum for
    freq_min : float
        minimum frequency of band for calculating noise
    freq_max : float
        maximum frequency of band for calculating noise
    
    Returns
    -------
    Dictionary with following items:
    freqs: frequencies of output spectrum
    spectrum: one-sided self-noise spectrum
    type: type of spectrum (power, amplitude, or dB, depends on spectrype input)
    spectrum_units: units of spectrum (depends on spectype input)
    freq_min: lower bound of noise band
    freq_max: upper bound of noise band
    noise: integrated noise over band
    noise_units: untis of integrated noise (depends on spectype input)

    Example
    -------
    ## calculate gem noise over the 0.1-20 Hz band as an RMS amplitude
    noise_info = gem_noise(version = '1.0', freq_min = 0.1, freq_max = 20, spectype = 'amp')
    print(noise_info['noise'])
    print(noise_info['noise_units'])
    
    """
    # use cases:
    # 1: wants output spectrum for given frequency vector
    # 2: wants output frequencies and spectrum with no frequency input
    # in order to permit 2, we need to return frequencies and spectrum
    if(float(version) >= 0.97):
        noise_spec = pd.read_csv(os.path.join(_noise_path, 'Gem_v0.98_Noise_spec.txt'))
    else:
        raise Exception('version %s not supported' % version)
    freq_in = noise_spec['f']
    power_in = noise_spec['amp']**2

    return _noise_spectrum_helper(freq_in, power_in, freq, spectype, freq_min, freq_max)

def ims_noise(model = 'low', freq = None, spectype = 'power', freq_min = None, freq_max = 50):
    """Return noise from an IMS noise model as a one-sided spectrum integrated
    over a defined frequency band

    Parameters
    ----------
    freq : numpy array
        Input frequencies for spectrum. If None, use default.
    spectype : str
        Type of spectrum and noise calculations: either 'power', 'amp', or 'dB' 
    freq_min : float
        minimum frequency of band for calculating noise
    freq_max : float
        maximum frequency of band for calculating noise
    
    Returns
    -------
    Dictionary with following items:
    freqs: frequencies of output spectrum
    spectrum: low noise model (one-sided self-noise spectrum)
    type: type of spectrum (power, amplitude, or dB, depends on spectrype input)
    spectrum_units: units of spectrum (depends on spectype input)
    freq_min: lower bound of noise band
    freq_max: upper bound of noise band
    noise: integrated noise over band
    noise_units: units of integrated noise (depends on spectype input)

    Example
    -------
    ## calculate gem noise over the 0.1-20 Hz band as an RMS amplitude
    noise_info = ims_noise(model = 'low', freq_min = 0.1, freq_max = 20, spectype = 'amp')
    print(noise_info['noise'])
    print(noise_info['noise_units'])
    
    """
    # use cases:
    # 1: wants output spectrum for given frequency vector
    # 2: wants output frequencies and spectrum with no frequency input
    # in order to permit 2, we need to return frequencies and spectrum
    noise_spec = pd.read_csv(os.path.join(_noise_path, 'IMSNOISE_MIN_MED_MAX.txt'),
                             names = ['f', 'min', 'med', 'max'], sep = r'\s+')
    freq_in = noise_spec['f']
    if model.lower() in ['low', 'min']:
        power_in = 10**noise_spec['min']
    elif model.lower() == 'med':
        power_in = 10**noise_spec['med']
    elif model.lower() in ['high', 'hi', 'max']:
        power_in = 10**noise_spec['max']
    else:
        raise Exception('Invalid IMS model: %s' % model)
    
    return _noise_spectrum_helper(freq_in, power_in, freq, spectype, freq_min, freq_max)



def _noise_spectrum_helper(freq_in, power_in, freq_out, spectype, freq_min, freq_max):
    if freq_out is None:
        freq_out = freq_in
    spec_function = scipy.interpolate.CubicSpline(freq_in, power_in, extrapolate = False)
    power = spec_function(freq_out)
    if (freq_min is not None) and (freq_max is not None):
        noise = scipy.integrate.quad(spec_function, freq_min, freq_max)[0] # 'quad' outputs are the estimated integral and error bar
    else:
        noise = np.NaN

    ## decide what kind of spectrum to return (default "power")
    if spectype.lower() == 'power':
        spec = power
        spec_units = 'Pa^2/Hz'
        noise_units = 'Pa^2'
    elif spectype.lower()[:3] == 'amp':
        spec = np.sqrt(power)
        noise = np.sqrt(noise)
        spec_units = 'Pa/sqrt(Hz)'
        noise_units = 'Pa'
    elif spectype.lower() == 'db':
        spec = 10*np.log10(power)
        noise = 10*np.log10(noise)
        spec_units = 'dB Pa/sqrt(Hz)'
        noise_units = 'dB Pa'
    else:
        raise Exception('spectype %s not supported' % spectype)

    return {'freqs':freq_out, 'spectrum':spec, 'type':spectype, 'spectrum_units':spec_units, 'freq_min':freq_min, 'freq_max':freq_max, 'noise':noise, 'noise_units':noise_units}


def _interpolate_stream(st, gap_limit_sec = 0.1):
    ## look for short data gaps and interpolate through them
    ## st must consist of data from just a single station
    st.merge()
    st = st.split()

    ## If there are no data gaps, the stream will only contain a single trace. Don't change anything.
    if len(st) == 1:
        return st
    
    for i in range(len(st) - 1):
        t1 = st[i].stats.endtime
        t2 = st[i+1].stats.starttime
        if (t2 - t1) < gap_limit_sec:
            ## define a new trace consisting of a masked array of the data gap and immediate surroundings
            st_new = st.slice(t1 - gap_limit_sec, t2 + gap_limit_sec)
            st_new.merge()
            tr_new = st_new[0]

            ## interpolate the missing values using a cubic spline
            data_good = ~tr_new.data.mask
            t = np.arange(len(tr_new.data))
            interp_function = scipy.interpolate.CubicSpline(t[data_good], tr_new.data[data_good], extrapolate = False)
            tr_new.data[~data_good] = interp_function(t[~data_good])

            ## add the new trace back to the stream
            st = st + tr_new
        else:
            ## if the data gap is not short, it will not be interpolated. For now, do nothing.
            pass
        
    st.merge() ## merge all the traces (done in place)
    st = st.split() ## if there are still data gaps, break them up (ensuring that the result is not a masked array)

    return st

            
            
            
    



__version__ = '1.5.14' # bug fix affecting edge case 
#__version__ = '1.5.13' # added executable and automated test for verify_huddle_test 
#__version__ = '1.5.12' # requires pandas<1.4 due to new conflict introduced in this version; fixed previous gemlog version name 
#__version__ = '1.5.11' # no longer supporting python 3.10 until the obspy-numpy conflict gets resolved. bug fixed with spurious data gaps/timing issues. 
#__version__ = '1.5.10' # more changes to support python 3.10 
#__version__ = '1.5.9' # improvements to installation process, cython now required 
#__version__ = '1.5.8' # new updates to support python 3.10 
#__version__ = '1.5.7' # support python 3.10 
#__version__ = '1.5.6' # added automated tests for minimum and newest versions to detect dependency incompatibilities 
#__version__ = '1.5.5' # allow deconvolve_gem_response to use non-standard response files, just like get_gem_response 
#__version__ = '1.5.4' # 'new checks for malformed files, added response files for older gems, test improvements' 
#__version__ = '1.5.3' # avoids timing problems caused by insufficient GPS lines in a file 
#__version__ = '1.5.2' # improved user interface to responses and noise specs 
#__version__ = '1.5.1' # added new version/platform info to gemconvert output and logs; improved elevation handling in metadata 
#__version__ = '1.4.3' # removed overlapping samples in output mseed files 
#__version__ = '1.4.2' # 'more flexibility in station_info.txt format' 
#__version__ = '1.4.1' # now uses default 24-hour miniSEED output files 
#__version__ = '1.4.0' # implemented timing corrections 
#__version__ = '1.3.13' # implemented cubic pps interpolation, but not timing correction (yet) 
#__version__ = '1.3.11' # fix bug in make_gem_inventory that affected surveys, where a sensor can move around and become multiple stations 
#__version__ = '1.3.10' # implement .wav file outputs 
#__version__ = '1.3.9' # still trying to get the pypi upload set right 
#__version__ = '1.3.8' # even more workflow fixes 
#__version__ = '1.3.7' # huddle_test improvements, minor change to make_db 
#__version__ = '1.3.6' # minor fix for inventory compliance with IRIS requirements 
#__version__ = '1.3.5' # updated gem_network, demo, and test so that inventory is written with elevation, start/end times for station/network, and azimuth/dip, satisfying IRIS requirements 
#__version__ = '1.3.4' # now supports format 0.91 (same, but with float pps-milliseconds for better clock drift estimation) 
#__version__ = '1.3.3' # bug fixes 
#__version__ = '1.3.2' # new code to merge day-long traces and change file naming format 
#__version__ = '1.3.1' # 'bug fix, and improvements to huddle test' 
#__version__ = '1.3.0' # improvements to gemconvert, added support for older raw file formats, and fixed bugs 
#__version__ = '1.2.7' # more bug fixes 
#__version__ = '1.2.6' # bug fixes to improve support for raw formats 0.8 and 0.85 and to make converting large datasets more reliable 
#__version__ = '1.2.5' # added support for raw format version 0.8 
#__version__ = '1.2.4' # bug fix in convert 
#__version__ = '1.2.3' # Changes to installation procedure (PyPI now available) and minor improvements to gemconvert 
#__version__ = '1.2.2' # debugging deployment code 
#__version__ = '1.2.1' # Run on mix of good raw files, no-gps raw files, and empty files without errors 
#__version__ = '1.2.0' # Handle missing GPS data more gracefully 
#__version__ = '1.1.2' # added output file length option to gemconvert, and fixed some missing_gps issues
#__version__ = '1.1.1' # moved drift correction from block-level to file-level
#__version__ = '1.1.0' # added new gem_cat feature for combining no-GPS files
#__version__ = '1.0.2' # changed make_gem_inventory so that output stationXML passes the stationXML validator: https://github.com/iris-edu/StationXML-Validator/
#__version__ = '1.0.1' # changed 'master' branch to 'main'
#__version__ = '1.0.0' # function name changes (some with aliases) to follow python standards
#__version__  = '0.3.2' # added demo with inventory functions
#__version__  = '0.3.1' #  more helpful logging output
#__version__  = '0.3.0' # 2020-09-04, cython added
#__version__  = '0.2.3' # improving testing, including empty/bad files
#__version__  = '0.2.2' # automated github tests, setup.py improvements, and modelst speed-up
#__version__  = '0.2.1' # 
#__version__ = '0.0.5' # added new functions to make network map from gps data and rename mseeds from serial_number.channel to network.station.location.channel codes
## List of old versions is not yet comprehensive##########################################



import numpy as np
import pandas as pd
import glob, obspy, os, warnings, gemlog
#from obspy.clients.nrl import NRL
#from contextlib import contextmanager,redirect_stderr,redirect_stdout
#from os import devnull
#nrl = NRL()

#response = nrl.get_response(sensor_keys = ['Gem', 'Gem Infrasound Sensor v1.0'], datalogger_keys = ['Gem', 'Gem Infrasound Logger v1.0', '0 - 128000 counts/V'])

#@contextmanager
#def _suppress_stdout_stderr():
#    """A context manager that redirects stdout and stderr to devnull"""
#    with open(devnull, 'w') as fnull:
#        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
#            yield (err, out)

def deconvolve_gem_response(data, gain = 'high', sensor_file = '', logger_file = ''):
    """
    Remove the Gem's instrument response
    
    Parameters:
    -----------
    data : either obspy.Stream or obspy.Trace
    Input data to remove response from. Note that if a stream is provided, the Gem response will be 
    deconvolved from all traces, even if some traces were not recorded by Gems!

    gain : str, default 'high'
    If a configuration file was used to set the Gem's programmable gain to half-gain, use 'low'.

    Returns:
    --------
    The input data (either trace or stream) with response removed
    
    Example:
    --------
    import obspy, gemlog
    ## read sample data--this isn't from a Gem, but it works
    tr = obspy.read()[0] 
    tr_deconvolved = gemlog.deconvolve_gem_response(tr)
    """
    ## obspy's tool to read instrument responses always triggers a warning for infrasound sensors
    ## use the warnings package to suppress warnings in this function only
    if type(data) == obspy.Stream:
        for tr in data:
            tr.stats.response = get_gem_response(gain, sensor_file, logger_file)
    elif type(data) == obspy.Trace:
        data.stats.response = get_gem_response(gain, sensor_file, logger_file)
    else:
        raise TypeError(f'data is type {type(data)}; must be obspy.Stream or obspy.Trace')

    ## By now, the response has been attached to all traces. Remove the response and return the data.
    data.remove_response()
    return data

## Create response variables. Do this once on package import so it doesn't have to be run repeatedly.
## obspy's tool to read instrument responses always triggers a warning for infrasound sensors
## use the warnings package to suppress warnings in this function only
#with warnings.catch_warnings():
#    warnings.simplefilter("ignore")
#_response_high_gain_1_0 = nrl.get_response(sensor_keys = ['Gem', 'Gem Infrasound Sensor v1.0'],
#                                           datalogger_keys = ['Gem', 'Gem Infrasound Logger v1.0',
#                                                              '0 - 128000 counts/V']) # may cause warning--ok to ignore
#_response_low_gain_1_0 =  nrl.get_response(sensor_keys = ['Gem', 'Gem Infrasound Sensor v1.0'],
#                                           datalogger_keys = ['Gem', 'Gem Infrasound Logger v1.0',
#                                                              '1 - 64000 counts/V']) # may cause warning--ok to ignore

def _read_response(filename):
    return obspy.read_inventory(filename)[0][0][0].response

def get_gem_response(gain = 'high', sensor_file = '', logger_file = ''):
    """
    Return the Gem's instrument response
    
    Parameters:
    -----------
    gain : str, default 'high'
    If a configuration file was used to set the Gem's programmable gain to half-gain, use 'low'.
    High gain is normal.

    sensor_file : str
    Not normally used; allows user to select a non-standard response (uncommon)

    logger_file : str
    Not normally used; allows user to select a non-standard response (uncommon)

    Returns:
    --------
    The Gem's instrument response as obspy.core.inventory.response.Response

    Example:
    --------
    import obspy, gemlog
    ## read sample data--this isn't from a Gem, but it works as a demo
    tr = obspy.read()[0] 

    ## note that the remainder of this example can be done more directly using 
    gemlog.deconvolve_gem_response

    ## find the Gem's response
    response = gemlog.get_gem_response()

    ## attach the response to the trace. 
    tr.stats.response = response

    ## remove the response from the trace in place
    tr.remove_response()
    
    """
    response_path = os.path.join(os.path.dirname(gemlog.__file__), 'data', 'response')
    if sensor_file == '':
        sensor_file = 'RESP.XX.IS025..BDF.GEMV1.26.0_0022'
    sensor_resp = _read_response(os.path.join(response_path, 'sensor', sensor_file))

    if logger_file == '':
        if gain.lower() == 'high':
            logger_file = 'RESP.XX.GM002..HHZ.GEMINFRAV1.0.100'
        elif gain.lower() == 'low':
            logger_file = 'RESP.XX.GM002..HHZ.GEMINFRAV1.0.100'
        else:
            raise ValueError(f'invalid gain: {gain}')
    response = _read_response(os.path.join(response_path, 'datalogger', logger_file))

    ## code copied from obspy's nrl.get_response() to merge sensor and logger responses
    response.response_stages.pop(0)
    sensor_stage0 = sensor_resp.response_stages[0]
    response.response_stages.insert(0, sensor_stage0)
    response.instrument_sensitivity.input_units = sensor_stage0.input_units
    response.instrument_sensitivity.input_units_description = sensor_stage0.input_units_description
    
    ## Obspy's get_response method fails to set the overall sensitivity correctly for infrasound
    ## responses, so we have to do this manually. If this isn't done, we'll get a warning later.
    ## We have to do this at freq 0.05 Hz for compatibility with the datalogger's nominal gain freq.
    filter_gain_0_05 = np.abs(1-1/(1+0.05j/0.039))
    response.instrument_sensitivity.value = response.response_stages[0].stage_gain * response.response_stages[2].stage_gain * filter_gain_0_05
    return response

## Two issues with obspy warnings with instrument responses:
## issue 1: verifying the response with nrl.get_response. After the sensor and datalogger responses are combined, it runs dl_resp.recalculate_overall_sensitivity() as a sanity check. Unfortunately, this method is unaware of pressure and pascals, and fails if given any option other than displacement, velocity, or acceleration. This can be suppressed with the warnings package, or avoided by just not using nrl.get_response().
## issue 2: removing the response, which ultimately calls response._call_eval_resp_for_frequencies(). This calls compiled C code, and the C code complains because it can't verify the sensitivity. This cannot be suppressed by either the warnings package or the io suppression. However, it can be suppressed by MANUALLY changing the overall instrument gain for the gem response, which is now done at the end of get_gem_response. Warnings should be expected for any other means of getting the gem response.
def _fix_station_info_keys(d):
    ## after reading a station_info file, standardize the key capitalization
    ## this saves the user from having to be picky about capitalizing the header right
    allowed_keys = ['SN', 'network', 'station', 'location', 'elevation']
    for key in allowed_keys:
        ## for the given allowed key, find the existing key that matches it
        w = [test_key.lower()[:2] == key.lower()[:2] for test_key in d.keys()]
        if any(w):
            ## rename that key to the standard name
            d[key] = d.pop(list(d.keys())[np.where(w)[0][0]])
    return d

def _get_station_info(station_info):
    required_keys = ['sn', 'network', 'station', 'location']
    if type(station_info) == str:
        print('Reading file %s' % station_info)
        header_df = pd.read_csv(station_info, nrows = 1, header = None, index_col = False)
        header_list = [header_df[key][0] for key in header_df.keys()]
        if all([i.lower() in (required_keys + ['elevation']) for i in header_list]): # file has header line
            print('File has valid header, using that for column names')
            station_info = pd.read_csv(station_info, dtype = {key:'str' for key in header_list}, keep_default_na = False)
        else:
            if len(header_list) == 4:
                print('File does not have a valid header, using default columns [SN, network, station, location]')
                station_info = pd.read_csv(station_info, names = required_keys, dtype = {key:'str' for key in header_list}, keep_default_na = False, index_col = False)
            elif len(header_list) == 5:
                print('File does not have a valid header, using default columns [SN, network, station, location, elevation]')
                station_info = pd.read_csv(station_info, names = required_keys + ['elevation'], dtype = {key:'str' for key in header_list}, keep_default_na = False, index_col = False)
            else:
                raise Exception('invalid station_info file; must have 4 or 5 columns or valid header')
        ## if any keys are capitalized or abbreviated wrong, correct them
        station_info = _fix_station_info_keys(station_info)
    elif (type(station_info) is not pd.DataFrame) or any([key not in station_info.keys() for key in required_keys]):
        raise Exception('invalid station_info input')
    # if location and network fields are blank in file, they are interpreted as NaN and must be
    # turned back into blank
    if 'elevation' not in station_info.keys():
        station_info['elevation'] = np.zeros(station_info.shape[0]) - 9999
        
    if station_info['elevation'].dtype == 'str':
        station_info['elevation'] = station_info['elevation'].astype('float')
        
    station_info.loc[station_info.location.isna(), 'location'] = ''
    station_info.loc[station_info.network.isna(), 'network'] = ''
    station_info.loc[station_info.elevation.isna(), 'elevation'] = -9999
    return station_info

def make_gem_inventory(station_info, coords, response = 'default'):
    """
    Create a station inventory for a Gem dataset.

    Parameters
    ----------
    station_info: str or pandas.DataFrame
        file that contains a table with station definitions (columns for serial number, network,
        station, location, and channel)
    
    coords : pandas.DataFrame
        output of summarize_gps(), with 'elevation' column added

    response: str
        instrument response information (currently only 'default' is supported)

    Returns
    -------
    obspy.Inventory containing station metadata for the dataset
        
    """
    ## ensure that the provided station_info has all the necessary info, and fail if it doesn't
    station_info = _get_station_info(station_info)
    ## ensure that the coords info has all the necessary info, and fail if it doesn't
    if (type(coords) is not pd.DataFrame):
        raise Exception('invalid coords; must be pandas.DataFrame')
    elif ('SN' not in coords.keys()) and any([key not in coords.keys() for key in ['network', 'station', 'location']]):
        raise Exception("invalid coords; must contain key 'SN', or keys 'network', 'station', 'location'")
    elif any([key not in coords.keys() for key in ['lat', 'lon', 'elevation']]):
        raise Exception("invalid coords; must contain keys 'lat', 'lon', 'elevation'")
    
    ## create the inventory and loop through all the networks, stations, and locations in it
    inventory = obspy.Inventory()
    all_networks = _unique(station_info['network'])
    for network_name in all_networks:
        network = obspy.core.inventory.Network(network_name)
        ## loop through all the stations in the current network
        all_stations = _unique(station_info[station_info['network'] == network_name]['station'])
        for station_name in all_stations:
            station = obspy.core.inventory.Station(station_name, latitude = 0, longitude = 0, elevation = 0)
            station.site.name = station_name
            ## loop through all the locations/channels in the current station
            all_locations = _unique(station_info[(station_info['network'] == network_name) & (station_info['station'] == station_name)]['location'])
            for location_name in all_locations:
                ## find the serial number for this location
                SN = station_info[(station_info['network'] == network_name) & (station_info['station'] == station_name) & (station_info['location'] == location_name)]['SN'].iloc[0]
                ## find the coords line corresponding to this location
                if ('SN' in coords.keys()) and any(key not in coords.keys() for key in ['network', 'station', 'location']):
                    line = coords[coords['SN'].astype(int) == int(SN)] # int just in case there's some type discrepancy
                else:
                    #line = coords[(coords['network'] == network_name) & (coords['station'] == station_name) & (coords['location'].astype(int) == int(location_name))] ## this crashes when location is ''
                    line = coords[(coords['network'] == network_name) & (coords['station'] == station_name) & (coords['location'] == location_name)]
                ## if no GPS info for this station is found, skip it
                if line.shape[0] == 0:
                    print('No coords found for %s.%s.%s, skipping' % (network_name, station_name, location_name))
                    continue
                ## We need to extract the coordinate and times for this location.
                lat = line['lat'].iloc[0]
                lon = line['lon'].iloc[0]
                elevation = line['elevation'].iloc[0]
                ## Check to see if the coords include start/end times. If so, pad
                ## them 1 hour to be safe. If not, assume the station is eternal.
                if all([key in coords.keys() for key in ['starttime', 'endtime']]):
                    t1 = obspy.UTCDateTime(line['starttime'].iloc[0]) - 3600 
                    t2 = obspy.UTCDateTime(line['endtime'].iloc[0]) + 3600
                else:
                    t1 = obspy.UTCDateTime('1970-01-01T00:00:00')
                    t2 = obspy.UTCDateTime('9999-12-31T23:59:59')
                ## Assume it's a gem v1.0--safe for now
                equipment = obspy.core.inventory.util.Equipment(serial_number = SN, model = 'Gem Infrasound Logger v1.0', description = 'Gem 1.0 (Infrasound), 0.039-27.1 Hz, 0.0035012 Pa/count')
                channel = obspy.core.inventory.Channel('HDF', location_code = location_name, latitude = lat, longitude = lon, elevation = elevation, depth = 0, response = response, equipments = equipment, start_date = t1, end_date = t2, sample_rate = 100, clock_drift_in_seconds_per_sample = 0, types = ['GEOPHYSICAL'], sensor = equipment, azimuth = 0, dip = 0)
                station.channels.append(channel)
            ## move on to the next station if this one is empty
            if len(station.channels) == 0:
                continue
            ## calculate the current station's coordinate as the mean of its locations,
            ## then append it to the network
            station.latitude = np.mean([channel.latitude for channel in station.channels])
            station.longitude = np.mean([channel.longitude for channel in station.channels])
            station.elevation = np.mean([channel.elevation for channel in station.channels])
            station.start_date = np.min([channel.start_date for channel in station.channels])
            station.end_date = np.max([channel.end_date for channel in station.channels])
            network.stations.append(station)
        ## append the current network to the inventory
        network.start_date = np.min([station.start_date for station in network.stations])
        network.end_date = np.max([station.end_date for station in network.stations])
        inventory.networks.append(network)
    return inventory
    
def rename_files(infile_pattern, station_info, output_dir, output_format = 'mseed', outfile_pattern =
                 '{year}-{mon}-{day}T{hour}_{min}_{sec}.{net}.{sta}.{loc}.{chan}.{fmt}'):
    """
    Rename a set of converted data files, assigning the correct network, station, and location 
    codes (instead of the original code with empty location/network and the station code being the
    serial number).

    Parameters
    ----------
    infile_pattern : str
        glob-type pattern defining the input files

    station_info : str or pandas.DataFrame
        file name for table assigning serial numbers to network, station, and location codes

    output_dir : str
        folder where output files should be written

    output_format : str
        default 'mseed'; 'sac' also works, as do other obspy-supported formats

    outfile_pattern : str
        format of output file names; note the abbreviations, and that 'jd' means 'day of year'

    Returns
    -------
    pandas.DataFrame containing the station_info table.
    """
    station_info = _get_station_info(station_info)
    
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir) # makedirs vs mkdir means if gpspath = dir1/dir2, and dir1 doesn't exist, that dir1 will be created and then di
    infiles = glob.glob(infile_pattern)
    infiles.sort()
    assert len(infiles) > 0, 'No input files provided'
    for i, infile in enumerate(infiles):
        st = obspy.read(infile)
        fileParts = infile.split('/')[-1].split('.')
        SN = fileParts[2]
        try:
            w = np.where(station_info.SN == SN)[0][0]
        except:
            print('skipping ' + infile)
            continue
        network = station_info.network[w]
        station = station_info.station[w]
        location = station_info.location[w]
        t1 = st[0].stats.starttime
        for tr in st:
            tr.stats.network = network
            tr.stats.station = station
            tr.stats.location = location
            t1 = min(t1, tr.stats.starttime)
        #outputFile = output_dir + '/' + '%s.%s.%s.%s.HDF.%s' % (fileParts[0], network, station, location, output_format)
        trace_info_dict = {'year':t1.year, 'mon':t1.month, 'day':t1.day,
                           'jd':t1.julday, 'hour':t1.hour, 'min':t1.minute,
                           'sec':t1.second, 'net':network, 'sta':station,
                           'loc':location, 'chan':'HDF', 'fmt':output_format}
        outfile_pattern = _fix_file_name_digits(outfile_pattern)
        output_file = outfile_pattern.format(**trace_info_dict)
        st.write(output_dir + '/' + output_file, format = output_format)
        print(str(i) + ' of ' + str(len(infiles)) + ': ' + infile + ', ' + output_file)
    return station_info

def merge_files_day(infile_path, infile_pattern = '*', outfile_dir = 'merge_file_output'):
    if not os.path.isdir(outfile_dir):
        os.makedirs(outfile_dir) # makedirs vs mkdir means if gpspath = dir1/dir2, and dir1 doesn't exist, that dir1 will be created and then dir2
    infiles = glob.glob(infile_path + '/' + infile_pattern)
    infiles.sort()
    ## find unique year, month, day lists for all files
    cuts = [infile.split('/')[-1].split('.')[0].split('T')[0] + ',' + '.'.join(infile.split('/')[-1].split('.')[1:])
            for infile in infiles]
    cuts = _unique(cuts)
    days = [cut.split(',')[0] for cut in cuts]
    suffixes = [cut.split(',')[1] for cut in cuts]
    
    ## for each item, read all the files and 
    for day, suffix in zip(days, suffixes):
        print(day + ' ' + suffix)
        x = obspy.read(infile_path + '/' + day + '*' + suffix)
        x.merge(fill_value = 'latest', method = 1)
        x.write(outfile_dir + '/' + day + 'T00_00_00.' + suffix)
    return

def _fix_file_name_digits(fn):
    fn = fn.replace('{year}', '{year:04d}')
    fn = fn.replace('{mon}', '{mon:02d}')
    fn = fn.replace('{day}', '{day:02d}')
    fn = fn.replace('{hour}', '{hour:02d}')
    fn = fn.replace('{min}', '{min:02d}')
    fn = fn.replace('{sec}', '{sec:02d}')
    fn = fn.replace('{jd}', '{jd:03d}')
    return fn


# function to get unique values from list while preserving order (set-based shortcut doesn't do this)
def _unique(list1): 
    unique_list = []
    for x in list1:
        # check if exists in unique_list or not 
        if x not in unique_list: 
            unique_list.append(x) 
    return unique_list

## function to exclude outliers by repeatedly calculating standard dev and tossing points outside N standard devs, until none are left
## this matters because occasionally a dataset will start or end with short recordings made elsewhere, which must be excluded from the calculation of station coords
def _remove_outliers(x, N = 5):
    w = np.where((np.abs(x.lat - np.median(x.lat)) < (N * (1e-5+np.std(x.lat)))) & (np.abs(x.lon - np.median(x.lon)) < (N * (1e-5+np.std(x.lon)))))[0]
    if len(w) < x.shape[0]:
        return _remove_outliers(x.iloc[w,:], N=N)
    else:
        return(x)

def read_gps(gps_dir_pattern, SN):
    """
    Read the most up-to-date GPS file for a given serial number.

    Parameters
    ----------
    gps_dir_pattern : str
        Path to the folder containing GPS data, or a glob-style pattern describing multiple folders.

    SN : str
        Serial number of the Gem being examined.

    Returns
    -------
    pandas.DataFrame containing columns year, date (day of year), lat, lon (all floats), and column
    t (obspy.UTCDateTime).
    """
    gpsDirList = sorted(glob.glob(gps_dir_pattern))
    gpsTable = pd.DataFrame(columns=['year', 'date', 'lat', 'lon', 't'])
    for gpsDir in gpsDirList:
        fnList = glob.glob(gpsDir + '/' + SN + '*')
        if len(fnList) > 0: # if any gps files matching SN are found, read and append the last
            gpsTable = gpsTable.append(pd.read_csv(sorted(fnList)[-1]), ignore_index=True)
    return gpsTable
ReadLoggerGPS = read_gps # alias; v1.0.0

def summarize_gps(gps_dir_pattern, station_info = None, output_file = None):
    """
    Read up-to-date GPS data from all Gems in a project, estimate their locations using a robust
    trimmed-mean method.

    Parameters
    ----------
    gps_dir_pattern : str
        Path to folder or glob-style pattern describing folders containing GPS data to review.

    station_info : str
        Path to text file containing table assigning serial numbers to network, station, and 
        location codes.

    output_file : str
        Path to file where output should be written (optional)

    Returns
    -------
    pandas.DataFrame containing the following columns:

        - SN: serial number (str)
        - lat: calculated average latitude (float)
        - lon: calculated average longitude (float)
        - lat_SE: standard error of average latitude (float)
        - lon_SE: standard error of average longitude (float)
        - starttime: time of first GPS data (obspy.UTCDateTime)
        - endtime: time of last GPS data (obspy.UTCDateTime)
        - num_samples: number of GPS strings recorded 

    If station_info is provided, then the following columns are also included:

        - network: network code (str)
        - station: station code (str)
        - location: location code (str)
        - elevation: elevation provided in station_info (str)
    """
    gpsDirList = sorted(glob.glob(gps_dir_pattern))
    gpsFileList = []
    for gpsDir in gpsDirList:
        gpsFileList += glob.glob(gpsDir + '/' + '*gps*txt')
    snList = []
    for gpsFile in gpsFileList:
        #snList.append(gpsFile.split('/')[-1][:3])
        snList.append(gpsFile.split('/')[-1].split('gps')[0])
    snList = sorted(_unique(snList))
    coords = pd.DataFrame(columns = ['SN', 'lat', 'lon', 'lat_SE', 'lon_SE', 'starttime', 'endtime', 'num_samples'])
    avgFun = lambda x: np.mean(x)
    seFun = lambda x: np.std(x)/np.sqrt(len(x))
    for i, SN in enumerate(snList):
        print(str(i) + ' of ' + str(len(snList)) + ': ' + SN)
        gpsTable = _remove_outliers(read_gps(gps_dir_pattern, SN))
        if gpsTable.shape[0] > 0:
            coords = coords.append(pd.DataFrame(
                [[SN, avgFun(gpsTable.lat), avgFun(gpsTable.lon), seFun(gpsTable.lat), seFun(gpsTable.lon), gpsTable.t.min(), gpsTable.t.max(), gpsTable.shape[0]]],
                columns = ['SN', 'lat', 'lon', 'lat_SE', 'lon_SE', 'starttime', 'endtime', 'num_samples']), ignore_index = True)
        else:
            print('No non-outliers remaining for gem {SN}')
    if station_info is not None:
        station_info = _get_station_info(station_info)
        network = []
        station = []
        location = []
        elevation = []
        for SN in coords.SN:
            try:
                w = np.where(station_info.SN == SN)[0][0]
                network.append(station_info.network[w])
                station.append(station_info.station[w])
                location.append(station_info.location[w])
                elevation.append(station_info.elevation[w])
            except:
                network.append('')
                station.append('')
                location.append('')
                elevation.append(-9999)
        coords['network'] = network
        coords['station'] = station
        coords['location'] = location
        coords['elevation'] = elevation
    if output_file is not None:
        coords.to_csv(output_file)
    return coords
    
SummarizeAllGPS = summarize_gps # alias, v1.0.0



import getopt
import pdb
import warnings
import numpy as np
from numpy import NaN, Inf
import os, glob, csv, time, scipy
import pandas as pd
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
import obspy
import sys
import shutil
from gemlog.gemlog import _read_single, EmptyRawFile, CorruptRawFile

def gem_cat(input_dir, output_dir, ext = '', cat_all = False):
    """
    gem_cat
    Merge raw data files so that all contain GPS data.
    
    Translated from R code originally by Danny Bowman.
    
    Parameters
    ----------
    input_dir: raw gem directory
    output_dir: path for renumbered and concatenated files
    ext: extension of raw files to convert (normally the serial number; sometimes TXT for old Gems)
    
    Returns
    -------
    None; file output only
    """
    if not os.path.isdir(input_dir):
        raise Exception('Input path ' + output_dir + ' does not exist')
    
    if not os.path.isdir(output_dir):
        try:
            os.makedirs(output_dir)
        except:
            raise Exception('Output path ' + output_dir + ' does not exist and cannot be created')
        else:
            print('Created output folder ' + output_dir)
    gem_files = sorted(glob.glob(input_dir + '/' + 'FILE[0-9][0-9][0-9][0-9].' + ext + '*'))
    
    has_gps = np.zeros(len(gem_files))
    counter = 0
    out_file = [] ## note that out_file being empty acts as a flag saying that no files have been processed yet
    try:
        SN = pd.read_csv(gem_files[0], delimiter = ',', skiprows = 4, nrows=1, dtype = 'str', names=['s', 'SN']).SN[0]
    except:
        raise EmptyRawFile(gem_files[0])

    for k in range(len(gem_files)):
        #for k in range(2):        
        print(str(k+1) + ' of ' + str(len(gem_files)) + ': ' + gem_files[k])

        if cat_all:
            if len(out_file) == 0:
                out_file = output_dir + '/FILE9999.' + SN
                shutil.copy(gem_files[k], out_file)
            else:
                AppendFile(gem_files[k], out_file, gem_files[k-1])
            continue

        try:
            lines = pd.read_csv(gem_files[k], delimiter = '\n', dtype = 'str', names = ['line'])
        except:
            raise EmptyRawFile(gem_files[k])
        #gps_grep = grepl("^G.*-?\\d+\\.\\d+,-?\\d+\\.\\d+$", lines)
        gps_grep = lines.line.str.contains('^G')
        ## if no files have been processed yet and the current file has no GPS data, skip it
   
        if (sum(gps_grep) == 0) & (len(out_file) == 0):
            continue

        ## if this is the first file being processed, simply copy it and advance
        if len(out_file) == 0:
            #out_file = output_dir + "/FILE", sprintf("%04d", counter) + "." + SN
            out_file = output_dir + '/FILE%04d.%s' % (counter, SN)
            shutil.copy(gem_files[k], out_file)
            counter += 1
            has_gps[k] = 1
            continue

        if sum(gps_grep) > 0:
            has_gps[k] = 1
            if has_gps[k - 1] == 0:
                ## if this isn't the first file being processed and it does have GPS data but the previous file didn't, append it to the current outfile
                AppendFile(gem_files[k], out_file, gem_files[k-1])
            else:
                ## if this isn't the first file being processed and it has gps data and the previous file did too, start a new outfile
                #out_file = output_dir + "/FILE" + sprintf("%04d", counter) + "." + SN
                #file_copy(gem_files[k], out_file, overwrite = TRUE)
                out_file = output_dir + '/FILE%04d.%s' % (counter, SN)
                shutil.copy(gem_files[k], out_file)
                counter += 1
        else:
            ## if this isn't the first file being processed and there's no GPS data, append it to the current outfile
            #if k == 1:
            #    next
            #else:
                ##system(paste0("cat ", gem_files[k], " >> ",    out_file))
            AppendFile(gem_files[k], out_file, gem_files[k-1])
    return 

#%%
#infile = gem_files[1]
#outfile = out_file
#prev_infile = gem_files[0]
#if True:
def AppendFile(infile, outfile, prev_infile):
    # ensure that the output path exists
    outpath = os.path.dirname(outfile)
    if not os.path.exists(outpath):
        os.makedirs(outpath)
        
    header = pd.read_csv(infile, delimiter = ',', nrows=1, dtype = 'str', names=['line']).line[0]
    format = header[7:]
    if format in ['0.85C', '0.9', '0.91']:
        #SN = scan(infile, skip = 4, sep = ',', what = list(character(), character()), nlines = 1, flush = TRUE)[[1]][2]
        ## determine what the last ADC reading is before the start of this file
        #if len(prev_infile) == 12:
        #    path = '.'
        #else:
        #    path = prev_infile[0:-12]
        ###########################################
        #L = suppressWarnings(ReadGem(nums = num, path = path, units = 'counts')) # suppressWarnings because it'll warn that there's no GPS issue (which is kind of the point) or SN (which doesn't matter)
        #p_start = L$p[length(L$p)]
        p_start = int(_read_single(prev_infile, require_gps = False, version = format)['data'][-1,1])
        ########################################

        ## read the first data line of the infile and convert it to an ADC reading difference
        j=0
        while True:
            ###################
            linetype = pd.read_csv(infile, skiprows = j, delimiter = '\n', nrows=1, dtype = 'str', names=['s']).s[0][0]
            ###############
            if linetype == 'D':
                break
            j += 1

        ## adjust the data line and append it to the outfile
        s = pd.read_csv(infile, delimiter = ',', nrows=1, skiprows=j, dtype = 'str', names=['c1','c2'])
        with open(outfile, 'a') as OF, open(infile, 'r') as IF:
            OF.write(s['c1'].iloc[0] + ',' + str(int(s['c2'].iloc[0])-p_start) + '\n')
            ## append the rest of the file
            #system(paste0('tail -n +', j+2, ' ', infile, ' >> ', outfile))
            for k, line in enumerate(IF):
                if k >= (j+1):
                    OF.write(line)
                #if k > 10: # testing only
                #    break
    else:
        ## earlier format version
        ## Find the end of the header and append the infile to the outfile past that point
        j=0
        while True:
            linetype = pd.read_csv(infile, skiprows = j, delimiter = '\n', nrows=1, dtype = 'str', names=['s']).s[0][0]
            if linetype == 'D':
                break
            j += 1
        
        ## append the rest of the file
        #system(paste0('tail -n +', j+1, ' ', infile, ' >> ', outfile))
        with open(outfile, 'a') as OF, open(infile, 'r') as IF:
            #system(paste0('tail -n +', j+2, ' ', infile, ' >> ', outfile))
            for k, line in enumerate(IF):
                if k >= j:
                    OF.write(line)

    return

def print_call():
    print('gem_cat -i <input_dir> -o <output_dir> -e <ext>')
    print('input_dir: raw gem directory')
    print('output_dir: path for renumbered and concatenated files')
    print('ext: extension of raw files to convert (normally the serial number; sometimes TXT for old Gems)')
    

def main(argv = None):
    if argv is None:
        argv = sys.argv[1:]
    inputdir = 'raw'
    outputdir = 'raw_merged'
    ext = ''
    try:
        opts, args = getopt.getopt(argv,"hi:o:e:")
    except getopt.GetoptError:
        print_call()
        sys.exit(2)
    for opt, arg in opts:
        if opt in ('-h', '--help'):
            print_call()
            sys.exit()
        elif opt in ("-i", "--inputdir"):
            inputdir = arg
        elif opt in ("-o", "--outputdir"):
            outputdir = arg
        elif opt in ("-e", "--ext"):
            ext = arg
    try:
        fn = os.listdir(inputdir)
    except:
        print("Problem opening input data folder " + inputdir + ". Did you give the right folder name after -i?")
        print("")
        print_call()
        sys.exit()
    if len(fn) == 0:
        print('Data folder ' + inputdir + ' does not contain any data files.')
        sys.exit()
    try:
        gem_cat(inputdir, outputdir, ext)
    except:
        print('gem_cat failed')
        sys.exit()
    else:
        print('Files merged successfully. Remember: sample times in the output are NOT precise!')
        

if __name__ == "__main__":
   main(sys.argv[1:])



"""
Cython-based file parsers.
"""

import numpy as np
from libc.stdio cimport fopen, fclose, fgets, FILE, sscanf


def parse_gemfile(filename):
    """
    Cythonized gem logfile parser.

    Parameters
    ----------
    filename : bytes
        The filename to parse. Must be of type `bytes` -- use
        filename.encode('utf-8') if needed.

    Returns
    -------
    tuple of three aligned numpy arrays:

        - 2-d array of numeric values read from the file
        - 1-d array of characters indicating the line type
        - 1-d array of the millisecond value of the row
    """
    cdef char* fname = filename

    cdef FILE* cfile
    cfile = fopen(fname, "rb")
    if cfile == NULL:
        msg = "No such file or directory: '{}'".format(filename)
        raise FileNotFoundError(2, msg)

    cdef char line[256]
    cdef char* read
    cdef char line_type = 0

    cdef int n_matched = 0
    # D placeholders
    cdef int ADC = 0
    cdef double DmsSamp = 0
    # G placeholders
    cdef int yr = 0, mo = 0, day = 0, hr = 0, mn = 0
    cdef double msPPS = 0, msLag = 0, sec = 0, lat = 0, lon = 0
    # M placeholders
    cdef int maxLag = 0, minFree = 0, maxUsed = 0, maxOver = 0
    cdef int gpsFlag = 0, freeStack1 = 0, freeStackIdle = 0
    cdef double ms = 0, batt = 0, temp = 0, A2 = 0, A3 = 0

    # array to store parsed data
    n_row = 780000  # max number of rows to expect: 750000 + 15000 + 15000
    result_array = np.zeros((n_row, 11), dtype=np.double)
    # make a view for faster indexing.
    # see https://cython.readthedocs.io/en/latest/src/userguide/numpy_tutorial.html#efficient-indexing-with-memoryviews
    cdef double[:, :] view = result_array

    # 1-D array to store linetype (single chars)
    result_linetypes = np.zeros(n_row, dtype='c')
    cdef char[:] type_view = result_linetypes
    # 1-D array to store millis.
    # range is 0 to 2**13, so choose short int
    result_millis = np.zeros(n_row, dtype=np.double)
    # cdef short[:] millis_view = result_millis
    cdef double[:] millis_view = result_millis

    cdef Py_ssize_t line_number = 0
    # were this python 3.8 we could maybe use the walrus operator.  alas
    while True:
        read = fgets(line, sizeof(line), cfile);
        if read == NULL:
            # EOF
            break

        line_type = line[0]
        if line_type == 68:  # ord('D') == 68
            # DmsSamp,ADC
            # D7780,-1
            n_matched = sscanf(line + 1, "%lf,%d", &DmsSamp, &ADC)
            view[line_number, 0] = ADC
            millis_view[line_number] = DmsSamp

        elif line_type == 71:  # ord('G') == 71
            # G,msPPS,msLag,yr,mo,day,hr,min,sec,lat,lon
            # G,8171,70,2020,6,20,5,21,22.0,43.62226,-116.20594
            n_matched = sscanf(line + 2,
                               "%lf,%lf,%d,%d,%d,%d,%d,%lf,%lf,%lf",
                               &msPPS, &msLag, &yr, &mo, &day, &hr, &mn,
                               &sec, &lat, &lon)
            view[line_number, 0] = msLag
            view[line_number, 1] = yr
            view[line_number, 2] = mo
            view[line_number, 3] = day
            view[line_number, 4] = hr
            view[line_number, 5] = mn
            view[line_number, 6] = sec
            view[line_number, 7] = lat
            view[line_number, 8] = lon
            millis_view[line_number] = msPPS

        elif line_type == 77:  # ord('M') == 77
            # M,ms,batt(V),temp(C),A2,A3,maxLag,minFree,maxUsed,maxOver,
            # gpsFlag,freeStack1,freeStackIdle
            # M,8001,3.02,22.1,1.412,2.052,94,66,9,0,0,57,86
            n_matched = sscanf(line + 2,
                               "%lf,%lf,%lf,%lf,%lf,%d,%d,%d,%d,%d,%d,%d",
                               &ms, &batt, &temp, &A2, &A3, &maxLag,
                               &minFree, &maxUsed, &maxOver, &gpsFlag,
                               &freeStack1, &freeStackIdle)
            view[line_number, 0] = batt
            view[line_number, 1] = temp
            view[line_number, 2] = A2
            view[line_number, 3] = A3
            view[line_number, 4] = maxLag
            view[line_number, 5] = minFree
            view[line_number, 6] = maxUsed
            view[line_number, 7] = maxOver
            view[line_number, 8] = gpsFlag
            view[line_number, 9] = freeStack1
            view[line_number, 10] = freeStackIdle
            millis_view[line_number] = ms
        else:
            continue

        type_view[line_number] = line_type
        line_number += 1

    fclose(cfile)

    return (
        result_array[:line_number, :],
        result_linetypes[:line_number],
        result_millis[:line_number],
    )



#/usr/bin/env python
# conda create -n g1 python=3.7.6 numpy obspy rpy2 pandas matplotlib
import sys # should always be available, doesn't need to be in "try"
try:
    import numpy as np
    import os, glob, getopt, logging, traceback, platform
    import gemlog
    from concurrent.futures import ProcessPoolExecutor
except Exception as e:
    print('Either dependencies are missing, or the environment is not active')
    print('Error message:')
    print(e)
    sys.exit(2)

#%%
def find_SN(x):
    return x[9:13]

# function to get unique values from list while preserving order (set-based shortcut doesn't do this)
def old_unique(list1):  # O(n^2), bad
    # intilize a null list 
    unique_list = [] 
    # traverse for all elements 
    for x in list1: 
        # check if exists in unique_list or not 
        if x not in unique_list: 
            unique_list.append(x) 
    return unique_list

def unique(list1): # thanks Kevin!
    unique, index = np.unique(list1, return_index=True)
    return sorted(unique)

def convert_single_SN(arg_list):
    inputdir, SN, outputdir, output_format, output_length = arg_list
    logging.info(f'Beginning {SN}')
    try:
        #print([inputdir, SN, outputdir, output_format, output_length])
        gemlog.convert(inputdir, SN = SN, convertedpath = outputdir, output_format = output_format, file_length_hour = output_length)
        print(f'{SN} done')
    except KeyboardInterrupt:
        logging.info('Interrupted by user')
        print('Interrupted')
        #sys.exit()
    except Exception as e:
        print(e)
        ####logging.info(traceback.format_exc(e.__traceback__))
        logging.exception(parse_error(e), exc_info = gemlog._debug)
        ####logging.exception(traceback.format_exc())
        logging.info('Error in ' + SN)
        pass
    else:
        pass
    logging.info('Completed ' + SN)
    return 0

def print_call():
    print('gemconvert -i <inputdir> -s <serialnumbers> -x <exclude_serialnumbers> -o <outputdir> -f <format> -l <filelength_hours> -p <number_of_processes>')
    print('-i --inputdir: default ./raw/')
    print('-s --serialnumbers: separate by commas (no spaces); default all')
    print('-x --exclude_serialnumbers: separate by commas (no spaces); default none')
    print('-o --outputdir: default ./mseed')
    print('-f --format: mseed, sac, or tspair (text) currently supported; default mseed')
    print('-l --length: length of output converted files in hours; default 24')
    print('-t --test: if used, print the files to convert, but do not actually run conversion')
    print('-p --parallel: number of processes to run in parallel (limited by your computer); default 1.')
    print('-h --help: print this message')
    print('alias: gem2ms. gemlog version: ' + gemlog.__version__)

def parse_error(e):
    e = str(e)
    if(e.find('Unable to allocate') > 0):
        return 'Tried to allocate very large data array, probably due to large time gap between adjacent files'
    elif(e.find('NULLType') > 0):
        return 'Suspected malformed file'
    else:
        return e
    

def main(argv = None):
    if argv is None:
        argv = sys.argv[1:]
    #print(sys.executable)
    inputdir = 'raw'
    SN_list = ''
    exclude = []
    outputdir = None
    test = False
    output_format = 'MSEED'
    output_length = 24 # hours
    num_processes = 1
    gemlog._debug = True
    try:
        opts, args = getopt.getopt(argv,"hdti:s:x:o:f:l:p:",["inputdir=","serialnumber="])
    except getopt.GetoptError:
        print_call()
        sys.exit(2)
    #print(2)
    #print(opts)
    for opt, arg in opts:
        #print([opt, arg])
        if opt in ('-h', '--help'):
            print_call()
            sys.exit()
        elif opt in ("-d", "--debug"):
            gemlog._debug = True
        elif opt in ("-i", "--inputdir"):
            inputdir = arg
        elif opt in ("-s", "--serialnumbers"):
            arg = arg.split(',')
            #print(arg)
            SN_list = arg
        elif opt in ("-x", "--exclude_serialnumbers"):
            arg = arg.split(',')
            #print(arg)
            exclude = arg
        elif opt in ("-t", "--test"):
            test = True
        elif opt in ("-o", "--outputdir"):
            outputdir = arg
        elif opt in ("-f", "--format"):
            output_format = arg
        elif opt in ("-l", "--length"):
            output_length = arg
        elif opt in ("-p", "--parallel"):
            try:
                num_processes = int(arg)
            except:
                'could not interpret number_of_processes ' + arg + ' as integer'

    if outputdir is None:
        outputdir = output_format.lower()
    if gemlog._debug:
        logging.basicConfig(level=logging.DEBUG, filename="gemconvert_logfile.txt", filemode="a+",
                        format="%(asctime)-15s %(levelname)-8s %(message)s")
    else:
        logging.basicConfig(level=logging.INFO, filename="gemconvert_logfile.txt", filemode="a+",
                        format="%(asctime)-15s %(levelname)-8s %(message)s")
        
    try:
        fn = os.listdir(inputdir)
    except:
        print("Problem opening input data folder " + inputdir + ". Did you give the right folder name after -i?")
        print("")
        print_call()
        sys.exit()

    if(len(SN_list) == 0): # if user does not provide SN_list, take unique SNs in order
        SN_list = unique([find_SN(fn[i]) for i in range(len(fn))]) # set takes unique values
        SN_list.sort()
    else: # if user provided SNs, keep the order, but take unique values
        SN_list = unique(SN_list)

    if (len(SN_list) == 0) & (len(opts) == 0):
        print("No data to convert; printing help message instead")
        print("")
        print_call()
        sys.exit()

    SN_list = [i for i in SN_list if i not in exclude]
    
    print(f'gemlog version {gemlog.__version__}')
    print('inputdir ', inputdir)
    print('serial numbers ', SN_list)
    print('outputdir ', outputdir)
    if not test:
        logging.info(f'***Starting conversion (gemlog version {gemlog.__version__})***')
        p = platform.uname()
        logging.info(f'Dependencies: Python {platform.python_version()}, obspy {gemlog.gemlog.obspy.__version__}, pandas {gemlog.gemlog.pd.__version__}, scipy {gemlog.gemlog.scipy.__version__}, numpy {gemlog.gemlog.np.__version__}')
        logging.info(f'platform.uname(): {p.system}, {p.release}, {p.version}, {p.machine}, {p.processor}')
        logging.info('Call: gemconvert ' + ' '.join(sys.argv[1:]))
        #for SN in SN_list:
        with ProcessPoolExecutor(max_workers=num_processes) as pool:
            args_list = [[inputdir, SN, outputdir, output_format, output_length] for SN in SN_list]
            res = pool.map(convert_single_SN, args_list)

if __name__ == "__main__":
   main(sys.argv[1:])



#!/usr/bin/env python
try:
    from setuptools import setup, Extension
except ImportError:
    raise RuntimeError('setuptools is required')

import os

from distutils.util import convert_path
import sys

# The minimum python version which can be used to run ObsPy
MIN_PYTHON_VERSION = (3, 7)

# Fail fast if the user is on an unsupported version of python.
if sys.version_info < MIN_PYTHON_VERSION:
    msg = ("gemlog requires python version >= {}".format(MIN_PYTHON_VERSION) +
           " you are using python version {}".format(sys.version_info))
    print(msg, file=sys.stderr)
    sys.exit(1)


DESCRIPTION = 'A set of functions for processing Gem datalogger files.'
LONG_DESCRIPTION = """
gemlog is a python package that provides functions for processing the
log files from the Gem infrasound datalogger. The Gem infrasound logger
is a low-cost, lightweight, low-power instrument for recording infrasound
in field campaigns.

Source code: https://github.com/ajakef/gemlog
"""

DISTNAME = 'gemlog'
LICENSE = 'GPL-3.0'
AUTHOR = 'Jake Anderson'
MAINTAINER_EMAIL = 'ajakef@gmail.com'
URL = 'https://github.com/ajakef/gemlog'


# TO-DO: replace this with versioneer
version_dict = {}
version_path = convert_path('gemlog/version.py')
with open(version_path) as version_file:
    exec(version_file.read(), version_dict)

VERSION = version_dict['__version__']

## example range: pandas>=1.3.0,<1.4.0
## example exact: numpy==1.21
INSTALL_REQUIRES = [
    'setuptools>=18.0', # 18.0 needed to handle cython in installation
    'obspy>=1.2.2', # June 2020, earlier versions don't install
    #'numpy>=1.17.3', # Oct 2019, required for pandas 1.3.0
    'numpy==1.21', # earlier numpy has security issue; 1.2.2 is incompatible with obspy 1.2.2. Make this >= once obspy updates.
    'pandas>=1.3.0,<1.4.0', # July 2021, earlier versions are incompatible with gemlog. 1.4.0 has new exceptions raised in read_csv that cause problems and are not addressed yet.
    'scipy>=1.3.0', # May 2019
    'matplotlib>=3.2.0', # March 2020
    'cython', # used for reading raw files quickly
    'fpdf' # needed for pdf huddle test reports
]

TESTS_REQUIRE = ['pytest']

EXTRAS_REQUIRE = {
    # 'optional': [...],
    'doc': ['sphinx==3.2.1', 'sphinx-rtd-theme==0.5.0'],
    'test': TESTS_REQUIRE
}
EXTRAS_REQUIRE['all'] = sorted(set(sum(EXTRAS_REQUIRE.values(), [])))

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Physics'
]

ENTRY_POINTS = {
    'console_scripts': [
        'gem2ms = gemlog.gem2ms:main',
        'gemconvert = gemlog.gem2ms:main',
        'gem_cat = gemlog.gem_cat:main',
        'verify_huddle_test = gemlog.huddle_test:main'
    ]
}

setuptools_kwargs = {
    'zip_safe': False,
    'scripts': [],
    'include_package_data': True,
}

PACKAGES = ['gemlog']

setup(name=DISTNAME,
      setup_requires=['setuptools>=18.0','cython'],
      version=VERSION,
      packages=PACKAGES,
      install_requires=INSTALL_REQUIRES,
      extras_require=EXTRAS_REQUIRE,
      tests_require=TESTS_REQUIRE,
      ext_modules=[Extension('gemlog.parsers', sources=['gemlog/parsers.pyx'])],
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      maintainer_email=MAINTAINER_EMAIL,
      license=LICENSE,
      url=URL,
      classifiers=CLASSIFIERS,
      entry_points=ENTRY_POINTS,
      **setuptools_kwargs)



from gemlog import parsers # cython file
from gemlog.gemlog import Convert, ReadGem, convert, read_gem, get_gem_specs
from gemlog.gem_network import summarize_gps, read_gps, SummarizeAllGPS, make_gem_inventory, rename_files, merge_files_day, get_gem_response, deconvolve_gem_response
from gemlog.gemlog_aux import make_db, calc_channel_stats, gem_noise, ims_noise
from gemlog.gem_cat import gem_cat
from gemlog.huddle_test import verify_huddle_test
from gemlog.version import __version__

#try:
#    from gemlog import parsers
#except ImportError:
#    print('Cannot find "parsers"; continuing')
#    pass
